"""State Command."""
from .common import create_ebus, disable_stdout_buffering


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("info", help="Show EBUSD meta information")
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    info = await ebus.async_get_info()
    for name, value in info.items():
        print(f"{name:<22s} {value}")



"""
Pythonic Interface to EBUS Daemon (EBUSD_).

Overview
========

* :any:`Ebus`: the EBUS handle, using one :any:`Connection` to an EBUSD instance.
  One EBUSD server can handle multiple :any:`Ebus` instances.
* :any:`MsgDef`: Message Definition containing multiple Field Defintions :any:`FieldDef`.
  A Virtual Field Definition :any:`VirtFieldDef` is a calculated value based on other fields.
* :any:`MsgDefs`: is a container for message definitions (:any:`MsgDef`).
* :any:`types`: contains the type engine, which allows the simple decode and
  encode of EBUSD values to/from python values.
* :any:`DummyConnection` emulates a connection. It answers requests by delegating to :any:`Dummy`.
* :any:`DummyServer` emulates an entire EBUS-daemon by delegating to :any:`Dummy`.
* :any:`Dummy` with :any:`DummyData` which emulates a minimal set of EBUS functions, for basic testing and trials.

All other classes and methods are just helper.

.. _EBUSD: https://github.com/john30/ebusd
"""
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

from . import types
from .circuitinfo import CircuitInfo
from .circuitmap import CircuitMap
from .connection import CommandError, Connection, Shutdown
from .const import AUTO, NA, OK
from .dummyconnection import DummyConnection
from .dummydata import DummyData
from .dummyserver import DummyServer
from .ebus import Ebus
from .exceptions import UnknownMsgError
from .icon import get_icon
from .msg import BrokenMsg, Field, Msg
from .msgdef import FieldDef, MsgDef, VirtFieldDef, resolve_prio
from .msgdefs import MsgDefs
from .prioritizer import Prioritizer



"""Dummy Server emulating EBUSD instance."""
import asyncio
import logging

from .dummy import Dummy, Quit, Stop

_LOGGER = logging.getLogger()


class DummyServer(Dummy):

    """
    Partial Dummy Server emulating EBUSD instance, just for testing.

    Keyword Args:
        port (int): Port. Default is 8888.
    """

    LOCALHOST = "127.0.0.1"

    def __init__(self, port=8888, dummydata=None):
        super().__init__(dummydata)
        self._port = port
        self._server, self._active = None, False

    def __repr__(self):
        return f"<DummyServer {self.LOCALHOST}:{self._port}>"

    @property
    def port(self):
        """Server Port."""
        return self._port

    async def async_start(self):
        """Start."""
        _LOGGER.info("%r: starting", self)

        async def _run():
            server = await asyncio.start_server(self, self.LOCALHOST, self.port)
            try:
                await server.serve_forever()
            except asyncio.CancelledError:
                pass

        self._active = True
        self._server = asyncio.create_task(_run())

    async def async_stop(self):
        """Stop."""
        if self._server:
            _LOGGER.info("%r: stopping", self)
            self._active = False
            self._server.cancel()

    async def async_wait_closed(self):
        """Wait closed."""
        if self._server:
            _LOGGER.info("%r: wait_closed", self)
            await self._server

    async def __call__(self, reader, writer):
        _LOGGER.info("%r: client connection started", self)
        while self._active:
            data = await reader.readline()
            try:
                for request in data.decode().split("\n"):
                    if request and self._active:
                        _LOGGER.info("%r: request: %r", self, request)
                        for response in self.respond(request):
                            _LOGGER.debug("%r: response: %r", self, response)
                            writer.write(f"{response}\n".encode())
                if not self._active:
                    _LOGGER.info("%r: shutdown", self)
                    writer.write("ERR: shutdown\n".encode())
            except Quit:
                break
            except Stop:
                await self.async_stop()
                break
        writer.close()
        _LOGGER.info("%r: client connection closed", self)



"""Pythonic EBUS Representation."""
import asyncio
import collections
import logging

from .circuitinfodecoder import decode_circuitinfos
from .connection import CommandError, Connection
from .const import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT, OK
from .exceptions import UnknownMsgError
from .msg import BrokenMsg, filter_msg
from .msgdecoder import MsgDecoder
from .msgdef import resolve_prio
from .msgdefdecoder import decode_msgdef
from .msgdefs import MsgDefs
from .util import repr_

_LOGGER = logging.getLogger(__name__)
_CMD_FINDMSGDEFS = "find -a -F type,circuit,name,fields"


class Ebus:

    """
    Pythonic EBUS Representation.

    The EBUS handle, using one :any:`Connection` to an EBUSD instance.
    One EBUSD server can handle multiple :any:`Ebus` instances.

    Keyword Args:
        host (str): EBUSD host
        port (int): EBUSD port
        timeout (int): Connection Timeout on connect and write.
        scaninterval (str): EBUSD scan - check interval
        scans (str): EBUSD scan - number of intervals
        circuitinfos (list): List with :any:`CircuitInfo` instances
        msgdefcodes (list): EBUSD Message Definition Codes
        msgdefs (MsgDefs): Message Definitions
    """

    # pylint: disable=R0902

    __slots__ = (
        "connection",
        "scaninterval",
        "scans",
        "msgdefcodes",
        "_msgdecoder",
        "_circuitinfos",
        "_circuitinfomap",
    )

    CONNECTOR = Connection
    DEFAULT_SCANINTERVAL = 10
    DEFAULT_SCANS = 3

    def __init__(
        self,
        host=DEFAULT_HOST,
        port=DEFAULT_PORT,
        timeout=DEFAULT_TIMEOUT,
        scaninterval=None,
        scans=None,
        circuitinfos=None,
        msgdefcodes=None,
        msgdefs=None,
    ):
        self._circuitinfomap = {}
        self.connection = self.CONNECTOR(host=host, port=port, autoconnect=True, timeout=timeout)
        self.scaninterval = scaninterval or self.DEFAULT_SCANINTERVAL
        self.scans = scans or self.DEFAULT_SCANS
        self.msgdefcodes = msgdefcodes or []
        self._msgdecoder = MsgDecoder(msgdefs or MsgDefs())
        self.circuitinfos = circuitinfos or []
        _LOGGER.info(repr(self))

    def __repr__(self):
        return repr_(
            self,
            kwargs=(
                ("host", self.host, DEFAULT_HOST),
                ("port", self.port, DEFAULT_PORT),
                ("timeout", self.timeout, DEFAULT_TIMEOUT),
                ("scaninterval", self.scaninterval, self.DEFAULT_SCANINTERVAL),
                ("scans", self.scans, self.DEFAULT_SCANS),
            ),
        )

    @property
    def ident(self):
        """Ident."""
        return f"{self.host}:{self.port}"

    @property
    def host(self):
        """Host Name or IP."""
        return self.connection.host

    @property
    def port(self):
        """Port."""
        return self.connection.port

    @property
    def timeout(self):
        """Timeout."""
        return self.connection.timeout

    @property
    def circuitinfos(self):
        """
        Circuit Informations :any:`CircuitInfo`.

        This property is writeable.
        """
        return self._circuitinfos

    @circuitinfos.setter
    def circuitinfos(self, circuitinfos):
        self._circuitinfos = tuple(circuitinfos)
        self._circuitinfomap = dict((circuitinfo.circuit, circuitinfo) for circuitinfo in self._circuitinfos)

    def get_circuitinfo(self, circuit):
        """Return :any:`CircuitInfo` for `circuit`."""
        try:
            return self._circuitinfomap[circuit]
        except KeyError:
            return None

    @property
    def msgdefs(self):
        """
        Message Defintions :any:`MsgDefs`.

        This property is writeable.."""
        return self._msgdecoder.msgdefs

    @msgdefs.setter
    def msgdefs(self, msgdefs):
        self._msgdecoder.msgdefs = msgdefs

    def __copy__(self):
        return Ebus(
            self.host,
            self.port,
            timeout=self.timeout,
            scaninterval=self.scaninterval,
            scans=self.scans,
            circuitinfos=self.circuitinfos,
            msgdefcodes=self.msgdefcodes,
            msgdefs=self.msgdefs,
        )

    async def async_wait_scancompleted(self):
        """
        Wait until EBUSD device scan is completed.

        EBUSD scans the bus infrastructure at startup be default.
        Devices and messages are detected during this time.
        This method waits until no new message are found.
        It checks every `scaninterval` seconds.
        The number of messages has to be stable for `scans` times.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        cnts = []
        while True:
            await self.connection.async_request(_CMD_FINDMSGDEFS)
            cnt = sum([1 async for line in self.connection.async_read()])
            cnts.append(cnt)
            if len(cnts) < self.scans or not all(cnt == cnts[-1] for cnt in cnts[-self.scans : -1]):
                await asyncio.sleep(self.scaninterval)
            else:  # pragma: no cover
                # not properly collected during coverage analysis
                break

    async def async_load_msgdefs(self):
        """
        Load Message Definitions from EBUSD.

        Alias for :any:`async_load_msgdefcodes` and :any:`decode_msgdefcodes`.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        await self.async_load_msgdefcodes()
        self.decode_msgdefcodes()

    async def async_load_msgdefcodes(self):
        """
        Load EBUS Message Definition Codes from EBUSD and store to :any:`msgdefcodes`.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.info("load_msgdefcodes()")
        self.msgdefcodes = msgdefcodes = []
        await self.connection.async_request(_CMD_FINDMSGDEFS)
        async for line in self.connection.async_read():
            line = line.strip()
            try:
                msgdef = decode_msgdef(line)
            except ValueError as exc:
                _LOGGER.warning("Cannot decode message definition %r (%s)", line, exc)
            if not msgdef.circuit.startswith("scan"):
                msgdefcodes.append(line)

    def decode_msgdefcodes(self):
        """Decode `msgdefcodes` and use as `msgdefs`."""
        _LOGGER.info("decode_msgdefcodes()")
        # Decode
        msgdefs = []
        for msgdefcode in self.msgdefcodes:
            try:
                msgdefs.append(decode_msgdef(msgdefcode))
            except ValueError as exc:
                _LOGGER.warning("Cannot decode message definition %r (%s)", msgdefcode, exc)
        # Sort
        self.msgdefs.clear()
        for msgdef in sorted(msgdefs, key=lambda msgdef: (msgdef.circuit, msgdef.name)):
            self.msgdefs.add(msgdef)

    async def async_read(self, msgdef, ttl=None, setprio=None):
        """
        Read Message.

        Args:
            msgdef (MsgDef): Message Definition

        Keyword Args:
            ttl (int): Time-to-live. Maximum age of read value in seconds.
            setprio: Priority `1-9` or `A` for automatic.

        Returns:
            Msg: Message

        Raises:
            ValueError: on decoder error
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.info("read(%r, ttl=%r), setprio=%r", msgdef, ttl, setprio)
        if setprio:
            msgdef = msgdef.replace(setprio=resolve_prio(msgdef, setprio))
        return await self._async_read(msgdef, ttl)

    async def async_write(self, msgdef, value, ttl=0):
        """
        Write Message.

        If `msgdef` just contains a subset of fields, the `value` is applied only to these, by
        running a read-modify-write operation.

        Args:
            msgdef (MsgDef): Message Definition

        Keyword Args:
            ttl (int): Time-to-live. Maximum age of read value in seconds.
                       Just needed in case of a Read-Modify-Write.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
        """
        _LOGGER.info("write(%r, value=%r, ttl=%r)", msgdef, value, ttl)
        if not msgdef.write:
            raise ValueError(f"Message is not writeable {msgdef}")
        fullmsgdef = self.msgdefs.get(msgdef.circuit, msgdef.name)
        if len(fullmsgdef.children) != len(msgdef.children):
            # Read
            if not msgdef.read:
                raise ValueError(f"Message is not read-modify-writable {msgdef}")
            await self.connection.async_request("read", msgdef.name, c=msgdef.circuit, m=ttl)
            line = await self.connection.async_readresp(check=False)
            values = line.split(";")
            # Modify
            for fielddef in msgdef.fields:
                encvalue = fielddef.type_.encode(value)
                values[fielddef.idx] = str(encvalue)
        else:
            values = [str(fielddef.type_.encode(value)) for fielddef in msgdef.fields]
        # Write
        await self.connection.async_request("write", msgdef.name, ";".join(values), c=msgdef.circuit)
        resp = await self.connection.async_readresp()
        if resp != "done":
            raise CommandError(resp)

    async def async_listen(self, msgdefs=None):
        """
        Listen to EBUS for messages.

        Listen to automatically updated messages and EBUSDs polling mechanism.

        Keyword Args:
            msgdefs (MsgDefs): Message definitions to be listened, other messages are ignored.

        Yields:
            Msg: Messages

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.info("listen(msgdefs=%r)", msgdefs)
        async for msg in self._async_listen(msgdefs):
            yield msg

    async def async_observe(self, msgdefs=None, ttl=None, setprio=None):
        """
        Observe `msgdefs` messages.

        Explicitly read all messages and then listen to automatically updated messages and
        EBUSDs polling mechanism.

        Keyword Args:
            msgdefs (MsgDefs): Message definitions to be observed, other messages are ignored.
            ttl (int): Time-to-live. Maximum age of read value in seconds.
            setprio: Priority `1-9` or `A` for automatic.

        Yields:
            Msg: Message

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.info("observe(msgdefs=%r, ttl=%r, setprio=%r)", msgdefs, ttl, setprio)
        msgdefs = msgdefs or self.msgdefs
        data = collections.defaultdict(lambda: None)

        # read all
        for msgdef in msgdefs:
            if msgdef.read:
                if setprio:
                    msgdef = msgdef.replace(setprio=resolve_prio(msgdef, setprio))
                msg = await self._async_read(msgdef, ttl=ttl)
                _LOGGER.debug("observe-read: %r", msg)
                msg = filter_msg(msg, msgdefs)
                if msg:
                    if msg.valid:
                        data[msgdef.ident] = msg
                    yield msg
            elif msgdef.update:
                data[msgdef.ident] = None

        # find new values (which got updated while we where reading)
        await self.connection.async_request("find -d")
        async for line in self.connection.async_read(check=False):
            msg = self._decode_msg(line)
            _LOGGER.debug("observe-find: %r", msg)
            msg = filter_msg(msg, msgdefs)
            if msg and msg != data[msg.msgdef.ident]:
                yield msg
                data[msg.msgdef.ident] = msg

        # listen
        async for msg in self._async_listen(msgdefs):
            _LOGGER.debug("observe-listen: %r", msg)
            yield msg

    async def async_get_state(self):
        """
        Return state string.

        This method does **NOT** raise any exception.

        Returns:
            str: OK or any error value.
        """
        _LOGGER.info("get_state()")
        return await self._async_get_state()

    async def _async_get_state(self):
        try:
            await self.connection.async_request("state")
            state = await self.connection.async_readresp(check=False)
            if state.startswith("signal acquired"):
                return OK
            else:
                return state
        except (ConnectionError, CommandError, ConnectionRefusedError):
            return "no ebusd connection"

    async def async_is_online(self):
        """
        Return `True` if everything is fine.

        This method does **NOT** raise any exception.

        Returns:
            bool
        """
        _LOGGER.info("is_online()")
        state = await self._async_get_state()
        return state == OK

    async def async_get_info(self):
        """
        Return EBUSD meta information.

        Returns
            dict: Meta information according to metainfo_

        .. _metainfo: https://github.com/john30/ebusd/wiki/3.1.-TCP-client-commands#info
        """
        _LOGGER.info("get_info()")
        info = {}
        await self.connection.async_request("info")
        async for line in self.connection.async_read():
            name, value = line.split(":", 1)
            info[name.strip()] = value.strip()
        return info

    async def async_load_circuitinfos(self):
        """Load EBUSD Circuit Information and store in :any:`circuitinfos`."""
        _LOGGER.info("load_circuitinfos()")
        await self.connection.async_request("info")
        lines = [line async for line in self.connection.async_read()]
        self.circuitinfos = decode_circuitinfos(lines)

    async def async_cmd(self, cmd, infinite=False, check=False):
        """
        Send EBUS_Command_ `cmd` to EBUSD and Receive Response.

        .. _EBUS_Command: https://github.com/john30/ebusd/wiki/3.1.-TCP-client-commands

        Args:
            cmd (str): Commmand String

        Keyword Args:
            infinite (bool): Do not stop at first empty line.
            check (bool): Abort on `ERR:` string in response.

        Yields:
            str: response

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: On connection breakdown.
            CommandError: If command failed and `check==True`.
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.info(f"cmd({cmd!r}, infinite={infinite!r}, check={check!r})")
        await self.connection.async_write(cmd)
        async for line in self.connection.async_read(infinite=infinite, check=check):
            yield line

    async def _async_read(self, msgdef, ttl=None):
        try:
            await self.connection.async_request("read", msgdef.name, c=msgdef.circuit, p=msgdef.setprio, m=ttl)
            line = await self.connection.async_readresp(check=False)
        except CommandError as exc:  # pragma: no cover
            return BrokenMsg(msgdef, str(exc))
        else:
            return self._msgdecoder.decode_value(msgdef, line)

    async def _async_listen(self, msgdefs):
        await self.connection.async_request("listen")
        resp = await self.connection.async_readresp()
        if resp != "listen started":
            raise CommandError(f"Listen could not be started: {resp}")
        async for line in self.connection.async_read(check=False):
            msg = self._decode_msg(line)
            msg = filter_msg(msg, msgdefs)
            if msg:
                yield msg

    def _decode_msg(self, line):
        if line:
            try:
                return self._msgdecoder.decode_line(line)
            except UnknownMsgError:
                pass
            except ValueError as exc:  # pragma: no cover
                _LOGGER.warning(f"Cannot decode message in {line!r}: {exc}")
        return None



"""
Message Definition Decoder.

EBUS message definitions are specified at
https://github.com/john30/ebusd/wiki/4.1.-Message-definition#message-definition .

The function :any:`decode_msgdef` converts a EBUS message defintion string into a :any:`MsgDef`.
"""
import collections
import re

from .msgdef import FieldDef, MsgDef
from .typedecoder import decode_type
from .types import EnumType
from .virtfielddef import iter_virtfielddefs


def decode_msgdef(line):
    """
    Decode Message and Field Definition retrieved from ebusd.

    The EBUSD command `find -a -F type,circuit,name,fields` retrieves
    all message and field definitions of all known and connected devices.
    The resulting lines are decoded by this method and create a proper
    :any:`MsgDef` instance per `line`.

    >>> m = decode_msgdef('r,mc.4,OtShutdownLimit,temp,s,UCH,,°C,"text, text"')
    >>> m.circuit, m.name, m.read, m.prio, m.write, m.update
    ('mc.4', 'OtShutdownLimit', True, None, False, False)
    >>> m.children
    (FieldDef(0, 'temp', IntType(0, 254), unit='°C', comment='text, text'),)

    >>> m = decode_msgdef('w,ui,TempIncrease,temp,m,D2C,,°C,Temperatur')
    >>> m.circuit, m.name, m.read, m.prio, m.write, m.update
    ('ui', 'TempIncrease', False, None, True, False)
    >>> m.children
    (FieldDef(0, 'temp', IntType(-2047.9, 2047.9, divider=16), unit='°C'),)
    """
    try:
        values = _split(line)
        type_, circuit, name = values[:3]  # pylint: disable=W0632
        read, prio, write, update = decodetype(type_)
        children = _decodefields(values[3:])
    except ValueError:
        raise ValueError(f"Invalid message definition {line!r}") from None
    for child in iter_virtfielddefs(children):
        children.append(child)
    return MsgDef(circuit, name, tuple(children), read, prio, write, update)


def _split(line):
    values = []
    regex = re.compile(r'("([^"]+)")|([^\,]*),')
    for mat in regex.finditer(line):
        groups = mat.groups()
        values.append(groups[1] or groups[2])
    return values


def decodetype(type_):
    """
    Decode Type.

    >>> decodetype('r')
    (True, None, False, False)
    >>> decodetype('w')
    (False, None, True, False)
    >>> decodetype('u')
    (False, None, False, True)
    """
    reg = re.compile(r"(r)([1-9]?)")
    mat = reg.match(type_)
    if mat:
        read = mat.group(1) is not None
        prio = int(mat.group(2)) if mat.group(2) else None
    else:
        read, prio = False, None
    write = "w" in type_
    update = not read and len(type_) > (1 if write else 0)
    return read, prio, write, update


def _decodefields(values):
    if len(values) % 6 in (0, 3, 4, 5):
        chunks = _chunks(values, 6)
        return list(_createfields(chunks))
    else:
        raise ValueError()


def _createfields(chunks):
    if chunks:
        # determine duplicate names
        fields = []
        dups = collections.defaultdict(lambda: -1)
        for chunk in chunks:
            name, _, datatype = chunk[:3]
            if not datatype.startswith("IGN"):
                fields.append(chunk)
                dups[name] += 1
        # create fields
        cnts = collections.defaultdict(lambda: 0)
        for idx, field in enumerate(fields):
            name = field[0]
            if dups[name]:
                cnt = cnts[name]
                cnts[name] = cnt + 1
                name = f"{name}.{cnt}"
            yield _createfield(idx, name, *field)


def _createfield(idx, name, _, __, datatype, dividervalues=None, unit=None, comment=None):
    if dividervalues and "=" in dividervalues:
        type_ = EnumType(tuple(pair.split("=", 1)[1] for pair in dividervalues.split(";")))
    else:
        ebustype = datatype.split(",")[0]
        if dividervalues:
            divider = float(dividervalues)
            if divider < 0:
                divider = 1 / -divider
        else:
            divider = None
        type_ = decode_type(ebustype, divider)
    return FieldDef(idx, name, type_, unit, comment)


def _chunks(list_or_tuple, maxsize):
    return [list_or_tuple[i : i + maxsize] for i in range(0, len(list_or_tuple), maxsize)]



"""State Command."""
from ..const import OK
from .common import create_ebus, disable_stdout_buffering


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("state", help="Show EBUSD state")
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    state = await ebus.async_get_state()
    if state != OK:
        raise Broken(state)
    print(state)


class Broken(RuntimeError):
    """Broken State"""



"""Mapping Of EBUS Circuit Names To Human-readable Names."""


class CircuitMap(dict):
    """
    Mapping of EBUS-circuit names To human-readable Names.

    >>> c = CircuitMap({
    ...     'broadcast': '*',
    ...     'bai': 'Heater',
    ...     'mc': 'Mixer',
    ...     'hwc': 'Water',
    ... })
    >>> for circuitname, humanname in c.items():
    ...     print(circuitname, '=', humanname)
    broadcast = *
    bai = Heater
    mc = Mixer
    hwc = Water
    >>> tuple(c)
    ('broadcast', 'bai', 'mc', 'hwc')

    Custom mappigns are added dictionary-like.

    >>> c = CircuitMap()
    >>> c['bai'] = 'Heater'
    >>> c['boo'] = 'My Boo'
    >>> c['mc.4'] = 'Mixer Unit 2'
    """

    def get_humanname(self, circuitname):
        """
        Return human-readable name for `circuitname`.

        The `.X` suffix distinguishes multiple instances of the same unit, this is handled gracefully.

        Returns:
            str: human-readable name

        >>> c = CircuitMap({
        ...     'broadcast': '*',
        ...     'bai': 'Heater',
        ...     'mc': 'Mixer',
        ...     'hwc': 'Water',
        ... })
        >>> c.get_humanname('bai')
        'Heater'
        >>> c.get_humanname('bai.3')
        'Heater#3'
        >>> c.get_humanname('mc.4')
        'Mixer#4'
        >>> c.get_humanname('unknown')
        'unknown'
        >>> c.get_humanname('unknown.4')
        'unknown.4'
        """
        # lookup full name
        humanname = self.get(circuitname, None)
        # loopup basename
        if humanname is None and "." in circuitname:
            basename, suffix = circuitname.split(".")
            humanname = self.get(basename, None)
            if humanname is not None:
                humanname = f"{humanname}#{suffix}"
        # use circuitname as default
        if humanname is None:
            humanname = circuitname
        return humanname



"""Generic Command."""

from .common import create_ebus


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser(
        "cmd",
        help=(
            "Issue TCP Command on EBUSD. "
            "See https://github.com/john30/ebusd/wiki/3.1.-TCP-client-commands for reference."
        ),
    )
    parser.add_argument(
        "--infinite", "-i", default=False, action="store_true", help="Do not abort command processing on empty line."
    )

    parser.add_argument(
        "cmd",
        help=("TCP Command. " "See https://github.com/john30/ebusd/wiki/3.1.-TCP-client-commands for reference."),
    )
    parser.set_defaults(main=_main)


async def _main(args):
    ebus = create_ebus(args)
    async for line in ebus.async_cmd(args.cmd, infinite=args.infinite, check=not args.infinite):
        print(line)



"""Common."""
import logging
import sys

import pyebus

_LOGGER = logging.getLogger(__name__).parent


def add_msgdef_args(parser):
    """Load arguments related to :any:`load_msgdef`."""
    parser.add_argument(
        "--scanwait",
        "-w",
        default=False,
        action="store_true",
        help=(
            "EBUSD scans the bus for available devices. "
            "Wait until this scan does not find any new messages. "
            "Specify this option, if EBUSD was started within the last minutes."
        ),
    )


def add_read_args(parser, ttl=None):
    """Read Arguments."""
    parser.add_argument("--defaultprio", "-p", default=None, help="Default priorty. 1-9 or A for AUTOMATIC.")
    parser.add_argument("--ttl", "-t", default=ttl, type=int, help="Maximum age of value in seconds.")


def add_patterns_arg(parser, opt=False):
    """Add patterns option."""
    if not opt:
        parser.add_argument("patterns", help="Message patterns separated by ';' (i.e. 'ui/OutsideTemp')")
    else:
        parser.add_argument(
            "patterns",
            nargs="?",
            default="*/*",
            help="Message patterns separated by ';' (i.e. 'ui/OutsideTemp'). Default is '*/*' for all.",
        )


def disable_stdout_buffering():
    """Disable STDOUT buffering."""
    sys.stdout = UnbufferedStream(sys.stdout)


def create_ebus(args):
    """Create :any:`Ebus` instance with parameters from `args`."""
    return pyebus.Ebus(host=args.host, port=args.port, timeout=args.timeout)


async def async_load_msgdefs(ebus, args):
    """Load Message Definitions."""
    if args.scanwait:
        print("Waiting for EBUSD scan to complete ... ", end="")
        await ebus.async_wait_scancompleted()
        print("DONE.")

    print("Loading Message Definitions ... ", end="")
    await ebus.async_load_msgdefs()
    print(f"{ebus.msgdefs.summary()} DONE.")


def print_msg(msg, skipbroken=False):
    """Formatted output."""
    if msg.valid:
        for field in msg.fields:
            # if isinstance(field.value, (ValueError))
            details = []
            fieldcomment = field.fielddef.comment
            if fieldcomment:
                details.append(f"({fieldcomment})")
            typecomment = field.fielddef.type_.comment
            if typecomment:
                details.append(f"{typecomment}")
            details = " ".join(details)
            unitvalue = field.unitvalue
            if unitvalue is None:
                unitvalue = ""
            print(f"{field.ident:<40s} {field.fielddef.msgdef.access} {unitvalue!s:<20s}{details}")
    elif not skipbroken:
        _LOGGER.error(f"{msg.msgdef.ident}: {msg.error}")


class UnbufferedStream:

    """Unbuffered `stream`."""

    def __init__(self, stream):
        self.stream = stream

    def write(self, *args, **kwargs):
        """Write."""
        self.stream.write(*args, **kwargs)
        self.stream.flush()

    def writelines(self, *args, **kwargs):
        """Write multiple lines."""
        self.stream.writelines(*args, **kwargs)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



"""Message Defintions."""
import collections
import re
from fnmatch import fnmatch

from .const import AUTO
from .msgdef import resolve_prio

_RE_RESOLVE = re.compile(r"\A([^/#+]+)/([^/#+]+)(#(\d|A))?(/([^/#]*))?\Z")


class Pattern(collections.namedtuple("Pattern", "circuit name setprio fieldname")):

    """Message Definition and Field Definition Search Pattern."""

    @staticmethod
    def from_str(pattern):
        """Create :any:`Pattern` from `pattern` string."""
        mat = _RE_RESOLVE.fullmatch(pattern)
        if mat:
            circuit, name, _, setprio, _, fieldname = mat.groups()
            return Pattern(circuit, name, setprio, fieldname)
        else:
            raise ValueError(f"Invalid pattern {pattern!r}")


class MsgDefs:

    """
    Message Definitions Container.

    >>> from .msgdef import MsgDef, FieldDef
    >>> from .types import Type
    >>> msgdefs = MsgDefs()
    >>> msgdefs.add(MsgDef('mc', 'Status0a', (
    ...     FieldDef(0, 'temp', Type(), '°C'),
    ...     FieldDef(1, 'mixer', Type(), None),
    ...     FieldDef(2, 'onoff-0', Type(), None),
    ...     FieldDef(3, 'onoff-1', Type(), None),
    ...     FieldDef(4, 'temp0', Type(), '°C'),
    ... ), read=True))
    >>> msgdefs.add(MsgDef('hc', 'Status0', (
    ...     FieldDef(0, 'temp', Type(), '°C'),
    ...     FieldDef(1, 'temp0', Type(), '°C'),
    ... ), read=True))
    >>> msgdefs.get('mc', 'Status0a')
    MsgDef('mc', 'Status0a', (FieldDef(0, 'temp', ...'°C'), FieldDef(1, 'mixer', ..., unit='°C')), read=True)
    >>> list(msgdefs)
    [MsgDef('mc', 'Status0a', (FieldDef(0, 'temp', ...'°C'), FieldDef(1, 'mixer', ..., unit='°C')), read=True)]
    >>> msgdefs.summary()
    '2 messages (2 read, 0 update, 0 write) with 7 fields'
    """

    def __init__(self):
        self.clear()

    def clear(self):
        """Remove All Stored Message Definitions."""
        self._msgdefs = collections.defaultdict(lambda: collections.defaultdict(list))

    def add(self, msgdef):
        """Add Message Definition."""
        msgdefs = self._msgdefs[msgdef.circuit][msgdef.name]
        for idx, msgdef0 in enumerate(msgdefs):
            joined = msgdef0.join(msgdef)
            if joined is not None:
                msgdefs[idx] = joined
                break
        else:
            msgdefs.append(msgdef)

    def get(self, circuit, name):
        """
        Get message with `circuit` and `name`.

        Returns
            MsgDef: Message Definition
        """
        msgdefs = self._msgdefs
        if circuit in msgdefs:
            circuitmsgdefs = msgdefs[circuit]
            if name in circuitmsgdefs:
                return circuitmsgdefs[name][0]
        return None

    def get_ident(self, ident):
        """
        Get message with `ident`.

        Returns
            MsgDef: Message Definition
        """
        pat = Pattern.from_str(ident)
        return self.get(pat.circuit, pat.name)

    def find(self, circuit, name="*"):
        """
        Find all message definitions matching `circuit` and `name`, wildcards and placeholder are accepted.

        Returns:
            MsgDefs: Message Definitions
        """
        msgdefs = MsgDefs()
        circuit = circuit.lower()
        name = name.lower()
        for msgdef in self:
            if fnmatch(msgdef.circuit.lower(), circuit) and fnmatch(msgdef.name.lower(), name):
                msgdefs.add(msgdef)
        return msgdefs

    def resolve(self, patterns, filter_=None):
        """
        Resolve patterns and filter message definitions.

        Returns:
            MsgDefs: Message Definitions
        """
        if isinstance(patterns, str):
            patterns = patterns.split(";")
        msgdefs = MsgDefs()
        for pattern in patterns:
            for msgdef in self._resolve(pattern.strip()):
                if msgdef not in msgdefs and (filter_ is None or filter_(msgdef)):
                    msgdefs.add(msgdef)
        return msgdefs

    def _resolve(self, pattern):
        pat = Pattern.from_str(pattern)
        circuit, name, setprio, fieldname = pat
        for msgdef in self.find(circuit, name):
            if fieldname is None:
                fields = msgdef.children
            else:
                fields = tuple(
                    fielddef for fielddef in msgdef.children if fnmatch(fielddef.name.lower(), fieldname.lower())
                )
            if not fields:
                continue
            if fields == msgdef.children and (setprio is None or not msgdef.read):
                yield msgdef
            else:
                setprio = setprio if setprio in (AUTO, None) else int(setprio)
                yield msgdef.replace(children=fields, setprio=setprio)

    def set_defaultprio(self, defaultprio):
        """Set Priorities of all messages without a priority value."""
        msgdefs = self._msgdefs
        for circuitmsgdefs in msgdefs.values():
            for msgdefs in circuitmsgdefs.values():
                for idx, msgdef in enumerate(msgdefs):
                    if msgdef.setprio is None:
                        msgdefs[idx] = msgdef.replace(setprio=resolve_prio(msgdef, msgdef.setprio or defaultprio))

    def summary(self):
        """
        Summary.

        Returns:
            str: Summary.
        """
        total = len(self)
        fields = sum(len(msgdef.children) for msgdef in self)
        read = sum([1 for msgdef in self if msgdef.read])
        update = sum([1 for msgdef in self if msgdef.update])
        write = sum([1 for msgdef in self if msgdef.write])
        return f"{total} messages ({read} read, {update} update, {write} write) with {fields} fields"

    def __iter__(self):
        for circuitmsgdefs in self._msgdefs.values():
            for msgdefs in circuitmsgdefs.values():
                yield from msgdefs

    def __len__(self):
        return sum(
            sum(len(msgdefs) for msgdefs in circuitmsgdefs.values()) for circuitmsgdefs in self._msgdefs.values()
        )

    def __add__(self, other):
        if self.__class__ is other.__class__:
            msgdefs = MsgDefs()
            for msgdef in self:
                msgdefs.add(msgdef)
            for msgdef in other:
                msgdefs.add(msgdef)
            return msgdefs
        else:
            return NotImplemented

    def __sub__(self, other):
        if self.__class__ is other.__class__:
            msgdefs = MsgDefs()
            othermsgs = tuple(other)
            for msgdef in self:
                if msgdef not in othermsgs:
                    msgdefs.add(msgdef)
            return msgdefs
        else:
            return NotImplemented



"""Type Engine."""
import datetime
import re

from .util import repr_


class Type:
    """Abstract Type."""

    def __init__(self):
        pass

    def __repr__(self):
        return repr_(self, self._getargs(), self._getkwargs())

    def _ident(self):
        return self._getargs(), self._getkwargs()

    def __hash__(self):
        return hash(self._ident())

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self._ident() == other._ident()
        else:
            return NotImplemented

    def _getargs(self):  # pylint: disable=R0201
        return tuple()

    def _getkwargs(self):  # pylint: disable=R0201
        return tuple()

    def with_divider(self, divider):
        """Return copy and apply `divider`."""
        raise NotImplementedError(self)

    def decode(self, value):
        """Decode `value`."""
        raise NotImplementedError(self)

    def encode(self, value):
        """Encode `value`."""
        raise NotImplementedError(self)

    @property
    def comment(self):
        """Get Comment on allowed values."""
        raise NotImplementedError(self)


class EnumType(Type):
    """
    Enumeration of `values`.

    Pure integer values are passed through, as they have an unknown coding.

    >>> t = EnumType(('on', 'auto', 'off'))
    >>> t
    EnumType(('on', 'auto', 'off'))
    >>> t.values
    ('on', 'auto', 'off')
    >>> t.comment
    'on, auto, off'

    >>> t.encode('auto')
    'auto'
    >>> t.encode(5)
    '5'
    >>> t.encode('5')
    '5'
    >>> t.encode('super')
    Traceback (most recent call last):
      ...
    ValueError: Unknown value 'super'. Allowed values are on, auto, off.
    >>> t.encode(None)
    '-'

    >>> t.decode('-') is None
    True
    >>> t.decode('auto')
    'auto'
    >>> t.decode('7')
    7
    >>> t.decode('super')
    Traceback (most recent call last):
      ...
    ValueError: Unknown value 'super'. Allowed values are on, auto, off.
    """

    _re_digit = re.compile(r"^\d+$")

    def __init__(self, values):
        super().__init__()
        self._values = values

    @property
    def values(self):
        """Enumeration Values."""
        return self._values

    def _getargs(self):
        return (self._values,)

    def decode(self, value):
        """Decode `value`."""
        if value == "-":
            value = None
        elif self._re_digit.match(value):
            value = int(value)
            # It does not matter to loop up the value, as it would have been found by ebusd
        elif value not in self._values:
            values = ", ".join(self._values)
            raise ValueError(f"Unknown value '{value}'. Allowed values are {values}.")
        return value

    def encode(self, value):
        """Encode `value`."""
        if value is None:
            value = "-"
        elif isinstance(value, int) or self._re_digit.match(str(value)):
            value = str(value)
        elif value not in self._values:
            values = ", ".join(self._values)
            raise ValueError(f"Unknown value '{value}'. Allowed values are {values}.")
        return value

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return ", ".join(self._values)


class StrType(Type):

    """
    String with optional maximum `length`.

    >>> t = StrType()
    >>> t
    StrType()
    >>> t.length is None
    True
    >>> t.encode('a string')
    'a string'
    >>> t.decode('a string')
    'a string'
    >>> t.comment is None
    True

    With length:

    >>> t = StrType(10)
    >>> t
    StrType(length=10)
    >>> t.length
    10
    >>> t.encode('a string')
    'a string'
    >>> t.decode('a string')
    'a string'
    >>> t.encode('a very long string')
    Traceback (most recent call last):
      ...
    ValueError: a very long string exceeds maximum length 10.
    >>> t.decode('a very long string')
    Traceback (most recent call last):
      ...
    ValueError: a very long string exceeds maximum length 10.
    >>> t.comment
    'Up to 10 characters'
    """

    def __init__(self, length=None):
        super().__init__()
        self._length = length

    @property
    def length(self):
        """Length."""
        return self._length

    def _getkwargs(self):
        return (("length", self.length, None),)

    def decode(self, value):
        """Decode `value`."""
        if self.length is not None and len(value) > self.length:
            raise ValueError(f"{value} exceeds maximum length {self.length}.")
        return value

    def encode(self, value):
        """Encode `value`."""
        if self.length is not None and len(value) > self.length:
            raise ValueError(f"{value} exceeds maximum length {self.length}.")
        return value

    @property
    def comment(self):
        """Get Comment on allowed values."""
        if self._length:
            return f"Up to {self._length} characters"
        else:
            return None


class HexType(Type):

    """
    Space Separated Hex Value with an optional `length` in number of Bytes.

    >>> t = HexType()
    >>> t.length is None
    True
    >>> t
    HexType()
    >>> t.decode('11 22 FF')
    (0x11, 0x22, 0xFF)
    >>> t.encode('0x11 0x22 0xFF')
    '11 22 FF'
    >>> t.encode('11 22 FF')
    '11 22 FF'
    >>> t.encode((0x11, 0x22, 0xFF))
    '11 22 FF'
    >>> t.comment
    'Hex Bytes'

    With length:

    >>> t = HexType(3)
    >>> t
    HexType(length=3)
    >>> t.length
    3
    >>> t.decode('11 22 FF')
    (0x11, 0x22, 0xFF)
    >>> t.decode('11 22')
    Traceback (most recent call last):
      ...
    ValueError: Hex value 11 22 has not expected length of 3
    >>> t.encode((0x11, 0x22, 0xFF))
    '11 22 FF'
    >>> t.encode((0x11, 0x22))
    Traceback (most recent call last):
      ...
    ValueError: Hex value (17, 34) has not expected length of 3
    >>> t.comment
    '3 Hex Bytes'
    """

    def __init__(self, length=None):
        super().__init__()
        self._length = length

    @property
    def length(self):
        """Width."""
        return self._length

    def _getkwargs(self):
        return (("length", self.length, None),)

    def decode(self, value):
        """Decode `value`."""
        values = value.split(" ")
        if self.length is not None:
            if len(values) != self.length:
                raise ValueError(f"Hex value {value} has not expected length of {self.length}")
        return tuple(Hex(int(value, 16)) for value in values)

    def encode(self, value):
        """Decode `value`."""
        if isinstance(value, str):
            value = [int(v, 16) for v in value.split(" ")]
        if self.length is not None:
            if len(value) != self.length:
                raise ValueError(f"Hex value {value} has not expected length of {self.length}")
        return " ".join((f"{v:02X}" for v in value))

    @property
    def comment(self):
        """Get Comment on allowed values."""
        if self._length:
            return f"{self._length} Hex Bytes"
        else:
            return "Hex Bytes"


class IntType(Type):

    """
    Integer in the range of [min_, max_] with granularity of `1 / divider`.

    >>> t = IntType(-4, 3)
    >>> t
    IntType(-4, 3)
    >>> t.min_
    -4
    >>> t.max_
    3
    >>> t.divider is None
    True
    >>> t.frac is None
    True
    >>> t.comment
    'Integer within [-4:3]'

    >>> t.decode('2')
    2
    >>> t.decode('-') is None
    True
    >>> t.decode('7')
    Traceback (most recent call last):
      ...
    ValueError: Value 7 exceeds upper limit of 3
    >>> t.decode('-5')
    Traceback (most recent call last):
      ...
    ValueError: Value -5 deceeds lower limit of -4

    >>> t.encode('2')
    2
    >>> t.encode(2)
    2
    >>> t.encode(None)
    '-'
    >>> t.encode(7)
    Traceback (most recent call last):
      ...
    ValueError: Value 7 exceeds upper limit of 3
    >>> t.encode(-5)
    Traceback (most recent call last):
      ...
    ValueError: Value -5 deceeds lower limit of -4

    With divider:

    >>> t = IntType(-4, 3, divider=2)
    >>> t
    IntType(-4, 3, divider=2)
    >>> t.min_
    -4
    >>> t.max_
    3
    >>> t.divider
    2
    >>> t.frac
    0.5
    >>> t.comment
    'Float within [-4:3] with 0.5 fraction'

    >>> t.decode('2')
    2.0
    >>> t.decode('-') is None
    True
    >>> t.decode('7')
    Traceback (most recent call last):
      ...
    ValueError: Value 7.0 exceeds upper limit of 3
    >>> t.decode('-5')
    Traceback (most recent call last):
      ...
    ValueError: Value -5.0 deceeds lower limit of -4

    >>> t.encode(2)
    2.0
    >>> t.encode(None)
    '-'
    >>> t.encode(7)
    Traceback (most recent call last):
      ...
    ValueError: Value 7.0 exceeds upper limit of 3
    >>> t.encode(-5)
    Traceback (most recent call last):
      ...
    ValueError: Value -5.0 deceeds lower limit of -4
    """

    def __init__(self, min_, max_, divider=None):
        super().__init__()
        self._min = min_
        self._max = max_
        self._divider = divider

    @property
    def min_(self):
        """Lower Limit."""
        return self._min

    @property
    def max_(self):
        """Upper Limit."""
        return self._max

    @property
    def divider(self):
        """Divider."""
        return self._divider

    @property
    def frac(self):
        """Fraction."""
        if self._divider and self._divider > 0:
            return 1 / self.divider
        else:
            return None

    def _getargs(self):
        return (self.min_, self.max_)

    def _getkwargs(self):
        return (("divider", self.divider, None),)

    def with_divider(self, divider):
        """Return copy and apply `divider`."""
        divider = _try_int(divider * (self.divider or 1))
        min_ = _try_int(self.min_ / divider)
        max_ = _try_int(self.max_ / divider)
        return IntType(min_, max_, divider=divider)

    def decode(self, value):
        """Decode `value`."""
        if value not in ("-", ""):
            if self.divider and self.divider > 0:
                value = float(value)
            else:
                value = int(value)
            if value < self.min_:
                raise ValueError(f"Value {value} deceeds lower limit of {self.min_}")
            if value > self.max_:
                raise ValueError(f"Value {value} exceeds upper limit of {self.max_}")
            return value
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if value is not None:
            if self.divider and self.divider > 0:
                value = float(value)
            else:
                value = int(value)
            if value < self.min_:
                raise ValueError(f"Value {value} deceeds lower limit of {self.min_}")
            if value > self.max_:
                raise ValueError(f"Value {value} exceeds upper limit of {self.max_}")
            return value
        else:
            return "-"

    @property
    def comment(self):
        """Get Comment on allowed values."""
        if self.divider and self.divider > 0:
            return f"Float within [{self.min_}:{self.max_}] with {self.frac} fraction"
        else:
            return f"Integer within [{self.min_}:{self.max_}]"


class BoolType(Type):

    """
    Boolean Type.

    >>> t = BoolType()
    >>> t
    BoolType()
    >>> t.comment
    '0 or 1'

    >>> t.decode('0')
    False
    >>> t.decode('1')
    True
    >>> t.decode('-') is None
    True

    >>> t.encode(0)
    0
    >>> t.encode(1)
    1
    >>> t.encode('0')
    0
    >>> t.encode('1')
    1
    >>> t.encode('false')
    0
    >>> t.encode('true')
    1
    >>> t.encode('FALSE')
    0
    >>> t.encode('TRUE')
    1
    >>> t.encode(None)
    '-'
    >>> t.encode('blub')
    Traceback (most recent call last):
        ...
    ValueError: blub is not a valid boolean
    """

    def decode(self, value):
        """Decode `value`."""
        if value not in ("-", ""):
            return bool(int(value))
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if value is not None:
            valuestr = str(value).lower()
            if valuestr in ("1", "true"):
                return 1
            elif valuestr in ("0", "false"):
                return 0
            else:
                raise ValueError(f"{value} is not a valid boolean")
        else:
            return "-"

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return "0 or 1"


class FloatType(Type):

    """
    Floating Type.

    >>> t = FloatType()
    >>> t
    FloatType()
    >>> t.comment
    'Float'
    >>> t.decode('-') is None
    True
    >>> t.decode('0')
    0.0
    >>> t.decode('1.456')
    1.456

    >>> t.encode(15.6)
    '15.6'
    >>> t.encode(None)
    '-'
    """

    def decode(self, value):
        """Decode `value`."""
        if value not in ("-", ""):
            value = float(value)
            return value
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if value is not None:
            value = float(value)
            return str(value)
        else:
            return "-"

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return "Float"


class DateType(Type):
    """
    Date Type.

    >>> import datetime
    >>> t = DateType()
    >>> t
    DateType()
    >>> t.comment
    'DAY.MONTH.YEAR'

    >>> t.decode('30.12.2020')
    datetime.date(2020, 12, 30)
    >>> t.decode('-.-.-') is None
    True

    >>> t.encode('30.12.2020')
    '30.12.2020'
    >>> t.encode(datetime.date(2020, 12, 30))
    '30.12.2020'
    >>> t.encode(None)
    '-.-.-'
    """

    _NONE = "-.-.-"

    def decode(self, value):
        """Decode `value`."""
        if value != self._NONE:
            return datetime.datetime.strptime(value, "%d.%m.%Y").date()
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%d.%m.%Y").date()
        if value is not None:
            return f"{value.day}.{value.month}.{value.year}"
        else:
            return self._NONE

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return "DAY.MONTH.YEAR"


class TimeType(Type):

    """
    Time.

    >>> t = TimeType()
    >>> t
    TimeType()
    >>> t.comment
    'HOUR:MINUTE:SECOND'
    >>> t.decode('23:59:59')
    Time(23, 59, 59)
    >>> t.decode('-:-:-') is None
    True
    >>> t.encode('23:59:59')
    '23:59:59'
    >>> t.encode(Time(23, 59, 59))
    '23:59:59'
    >>> t.encode(None)
    '-:-:-'
    """

    _NONE = "-:-:-"

    def decode(self, value):
        """Decode `value`."""
        if value != self._NONE:
            tstamp = datetime.datetime.strptime(value, "%H:%M:%S")
            return Time(tstamp.hour, tstamp.minute, tstamp.second)
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%H:%M:%S")
        if value is not None:
            return f"{value.hour:02d}:{value.minute:02d}:{value.second:02d}"
        else:
            return self._NONE

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return "HOUR:MINUTE:SECOND"


class HourMinuteType(TimeType):

    """
    Time.

    Keyword Args:
        minres: Minute Resolution

    >>> t = HourMinuteType()
    >>> t
    HourMinuteType()
    >>> t.minres is None
    True
    >>> t.comment
    'HOUR:MINUTE'
    >>> t.decode('23:59')
    HourMinute(23, 59)
    >>> t.decode('-:-') is None
    True
    >>> t.encode('23:59')
    '23:59'
    >>> t.encode(HourMinute(23, 59))
    '23:59'
    >>> t.encode(None)
    '-:-'

    With Minimum Resolution

    >>> t = HourMinuteType(minres=10)
    >>> t
    HourMinuteType(minres=10)
    >>> t.minres
    10
    >>> t.comment
    'HOUR:MINUTE with 10min granularity'
    >>> t.decode('23:50')
    HourMinute(23, 50)
    >>> t.decode('-:-') is None
    True
    >>> t.decode('23:51')
    Traceback (most recent call last):
      ...
    ValueError: Minute of 23:51 must be multiple of 10
    >>> t.encode('23:50')
    '23:50'
    >>> t.encode(HourMinute(23, 50))
    '23:50'
    >>> t.encode(HourMinute(23, 52))
    Traceback (most recent call last):
      ...
    ValueError: Minute of 23:52 must be multiple of 10
    >>> t.encode(None)
    '-:-'
    """

    _NONE = "-:-"

    def __init__(self, minres=None):
        super().__init__()
        self._minres = minres

    @property
    def minres(self):
        """Minute Resolution."""
        return self._minres

    def _getkwargs(self):
        return (("minres", self._minres, None),)

    def decode(self, value):
        """Decode `value`."""
        if value != self._NONE:
            value = datetime.datetime.strptime(value, "%H:%M")
            value = HourMinute(value.hour, value.minute)
            if self._minres and (value.minute % self._minres) != 0:
                raise ValueError(f"Minute of {value} must be multiple of {self._minres}")
            return value
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%H:%M")
        if value is not None:
            if self._minres and (value.minute % self._minres) != 0:
                raise ValueError(f"Minute of {value} must be multiple of {self._minres}")
            return f"{value.hour:02d}:{value.minute:02d}"
        else:
            return self._NONE

    @property
    def comment(self):
        """Get Comment on allowed values."""
        if self.minres:
            return f"HOUR:MINUTE with {self._minres}min granularity"
        else:
            return "HOUR:MINUTE"


class DateTimeType(Type):

    """
    Date Time.


    >>> t = DateTimeType()
    >>> t
    DateTimeType()
    >>> t.comment
    'DAY.MONTH.YEAR HOUR:MINUTE:SECOND'
    >>> t.decode('30.12.2020 23:59:59')
    DateTime(2020, 12, 30, 23, 59, 59)
    >>> t.decode('-.-.- -:-:-') is None
    True
    >>> t.encode('30.12.2020 23:59:59')
    '30.12.2020 23:59:59'
    >>> t.encode(DateTime(2020, 12, 30, 23, 59, 59))
    '30.12.2020 23:59:59'
    >>> t.encode(None)
    '-.-.- -:-:-'
    """

    _NONE = "-.-.- -:-:-"

    def decode(self, value):
        """Decode `value`."""
        if value != self._NONE:
            return DateTime.strptime(value, "%d.%m.%Y %H:%M:%S")
        else:
            return None

    def encode(self, value):
        """Encode `value`."""
        if isinstance(value, str):
            value = DateTime.strptime(value, "%d.%m.%Y %H:%M:%S")
        if value is not None:
            return f"{value.day}.{value.month}.{value.year} {value.hour:02d}:{value.minute:02d}:{value.second:02d}"
        else:
            return self._NONE

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return "DAY.MONTH.YEAR HOUR:MINUTE:SECOND"


class WeekdayType(Type):
    """
    Weekday Type.

    >>> t = WeekdayType()
    >>> t.comment
    ''
    >>> t.decode('a')
    'a'
    >>> t.encode('a')
    'a'
    """

    def decode(self, value):
        """Decode `value`."""
        return value

    def encode(self, value):
        """Encode `value`."""
        return value

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return ""


class PinType(Type):

    """
    Pin.

    >>> t = PinType()
    >>> t.comment
    ''
    >>> t.decode('a')
    'a'
    >>> t.encode('a')
    'a'
    """

    def decode(self, value):
        """Decode `value`."""
        return value

    def encode(self, value):
        """Encode `value`."""
        return value

    @property
    def comment(self):
        """Get Comment on allowed values."""
        return ""


class DateTime(datetime.datetime):

    """
    DateTime.

    >>> t = DateTime(2020, 12, 31, 23, 59, 59)
    >>> t
    DateTime(2020, 12, 31, 23, 59, 59)
    >>> str(t)
    '31.12.2020 23:59:59'
    """

    def __str__(self):
        return self.strftime("%d.%m.%Y %H:%M:%S")


class Time(datetime.time):

    """
    Time.

    >>> t = Time(23, 59, 59)
    >>> t
    Time(23, 59, 59)
    >>> str(t)
    '23:59:59'
    """

    def __str__(self):
        return self.strftime("%H:%M:%S")


class HourMinute(datetime.time):

    """
    Time without Seconds.

    >>> t = HourMinute(23, 59)
    >>> t
    HourMinute(23, 59)
    >>> str(t)
    '23:59'
    """

    def __str__(self):
        return self.strftime("%H:%M")


class Hex(int):

    """Integer with Hex Representation."""

    def __repr__(self):
        return f"0x{self:02X}"

    __str__ = __repr__


def _try_int(value):
    intvalue = int(value)
    if float(value) == float(intvalue):
        value = intvalue
    return value



"""Read Command."""
from .common import (
    add_msgdef_args,
    add_patterns_arg,
    add_read_args,
    async_load_msgdefs,
    create_ebus,
    disable_stdout_buffering,
    print_msg,
)


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("read", help="Read values from the bus, decode and print")
    add_msgdef_args(parser)
    add_read_args(parser, ttl=0)
    add_patterns_arg(parser, opt=True)
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    await async_load_msgdefs(ebus, args)
    msgdefs = ebus.msgdefs.resolve(args.patterns, filter_=lambda msgdef: msgdef.read or msgdef.update)
    if args.defaultprio:
        msgdefs.set_defaultprio(args.defaultprio)
    print(f"Reading to {msgdefs.summary()}")
    for msgdef in msgdefs:
        if msgdef.read:
            msg = await ebus.async_read(msgdef, ttl=args.ttl)
            print_msg(msg)



"""Exceptions."""


class UnknownMsgError(RuntimeError):

    """Exception raised in case of unknown Message."""


class CommandError(RuntimeError):

    """Command Error raised in case of EBUSD error, typically if the response starts with `ERR:`"""


class Shutdown(ConnectionError):

    """EBUS Shutdown."""



"""MDI Icon Utility."""
from . import types


def get_icon(fielddef, state=None):
    """
    Return Appropriate mdi-icon_ for `fielddef` at `state`.

    .. _mdi-icon: https://materialdesignicons.com/


    >>> from pyebus import FieldDef, types
    >>> get_icon(FieldDef(0, 'name', types.FloatType(), unit='°C'))
    'mdi:thermometer'
    >>> get_icon(FieldDef(0, 'name', types.TimeType()))
    'mdi:timer'
    >>> get_icon(FieldDef(0, 'name', types.EnumType(('on', 'off'))))
    'mdi:toggle-switch'
    >>> get_icon(FieldDef(0, 'name', types.EnumType(('on', 'off'))), 'off')
    'mdi:toggle-switch-off'
    >>> get_icon(FieldDef(0, 'name', types.EnumType(('on', 'off'))), 'on')
    'mdi:toggle-switch'
    """
    type_ = fielddef.type_
    if fielddef.unit in ("°C", "K", "°F"):
        return "mdi:thermometer"
    elif isinstance(type_, (types.TimeType, types.DateType, types.DateTimeType, types.HourMinuteType)):
        return "mdi:timer"
    elif isinstance(type_, types.EnumType):
        if tuple(sorted(type_.values)) in [("off", "on"), ("no", "yes")]:
            if state in (False, "off", "no"):
                return "mdi:toggle-switch-off"
            else:
                return "mdi:toggle-switch"
    return None



"""EBUSD Typecode Decoder."""
import re

from . import types

_RE_BIT = re.compile(r"\ABI\d(:(\d))?\Z")

TYPEMAP = {
    # BDA       BCD date                      dd.mm.yyyy               day first, including weekday, Sunday=0x06
    # BDA:3     BCD date                      dd.mm.yyyy               day first, excluding weekday
    # HDA       hex date                      dd.mm.yyyy               day first, including weekday, Sunday=0x07
    # HDA:3     hex date                      dd.mm.yyyy               day first, excluding weekday
    # DAY       date in days                  dd.mm.yyyy               days since 01.01.1900
    "BDA": types.DateType(),
    "BDA:3": types.DateType(),
    "HDA": types.DateType(),
    "HDA:3": types.DateType(),
    "DAY": types.DateType(),
    # BTI       BCD time                      hh:mm:ss                 seconds first
    # HTI       hex time                      hh:mm:ss                 hours first
    # VTI       hex time                      hh:mm:ss                 seconds first
    "BTI": types.TimeType(),
    "HTI": types.TimeType(),
    "VTI": types.TimeType(),
    # BTM       BCD time                      hh:mm                    minutes first
    # HTM       hex time                      hh:mm                    hours first
    # VTM       hex time                      hh:mm                    minutes first
    # MIN       time in minutes               hh:mm                    minutes since last midnight
    # TTM       truncated time                hh:m0                    multiple of 10 minutes
    # TTH       truncated time                hh:m0                    multiple of 30 minutes
    # TTQ       truncated time                hh:mm                    multiple of 15 minutes
    "BTM": types.HourMinuteType(),
    "HTM": types.HourMinuteType(),
    "VTM": types.HourMinuteType(),
    "MIN": types.HourMinuteType(),
    "TTM": types.HourMinuteType(minres=10),
    "TTH": types.HourMinuteType(minres=30),
    "TTQ": types.HourMinuteType(minres=15),
    # BDY       weekday                       Mon...Sun                Sunday=0x06
    # HDY       weekday                       Mon...Sun                Sunday=0x07
    "BDY": types.WeekdayType(),
    "HDY": types.WeekdayType(),
    # BCD       unsigned BCD                  0...99
    # BCD:2     unsigned BCD                  0...9999
    # BCD:3     unsigned BCD                  0...999999
    # BCD:4     unsigned BCD                  0...99999999
    "BCD": types.IntType(0, 99),
    "BCD:2": types.IntType(0, 9999),
    "BCD:3": types.IntType(0, 999999),
    "BCD:4": types.IntType(0, 99999999),
    # TODO: HCD       unsigned hex BCD              0...99999999             each BCD byte converted to hex
    # TODO: HCD:1     unsigned hex BCD              0...99                   each BCD byte converted to hex
    # TODO: HCD:2     unsigned hex BCD              0...9999                 each BCD byte converted to hex
    # TODO: HCD:3     unsigned hex BCD              0...999999               each BCD byte converted to hex
    # PIN       unsigned BCD                  0000...9999
    "PIN": types.PinType(),
    # UCH       unsigned integer              0...254
    "UCH": types.IntType(0, 254),
    # SCH       signed integer               -127...127
    # D1B       signed integer               -127...127
    "SCH": types.IntType(-127, 127),
    "D1B": types.IntType(-127, 127),
    # D1C       unsigned number               0.0...100.0              fraction 1/2 = divisor 2
    "D1C": types.IntType(0, 100, divider=2),
    # D2B       signed number                -127.99...127.99          fraction 1/256 = divisor 256
    "D2B": types.IntType(-127.99, 127.99, divider=256),
    # D2C       signed number                -2047.9...2047.9          fraction 1/16 = divisor 16
    "D2C": types.IntType(-2047.9, 2047.9, divider=16),
    # FLT       signed number                -32.767...32.767         low byte first, fraction 1/1000 = divisor 1000
    # FLR       signed number reverse        -32.767...32.767         high byte first, fraction 1/1000 = divisor 1000
    "FLT": types.IntType(-32.767, 32.767, divider=1000),
    "FLR": types.IntType(-32.767, 32.767, divider=1000),
    # EXP       signed float number          -3.0e38...3.0e38          low byte first
    # EXR       signed float number reverse  -3.0e38...3.0e38          high byte first
    "EXP": types.FloatType(),
    "EXR": types.FloatType(),
    # UIN       unsigned integer              0...65534                low byte first
    # UIR       unsigned integer reverse      0...65534                high byte first
    "UIN": types.IntType(0, 65534),
    "UIR": types.IntType(0, 65534),
    # SIN       signed integer               -32767...32767            low byte first
    # SIR       signed integer reverse       -32767...32767            high byte first
    "SIN": types.IntType(-32767, 32767),
    "SIR": types.IntType(-32767, 32767),
    # U3N       unsigned 3 byte int           0...16777214             low byte first
    # U3R       unsigned 3 byte int reverse   0...16777214             high byte first
    "U3N": types.IntType(0, 16777214),
    "U3R": types.IntType(0, 16777214),
    # S3N       signed 3 byte int            -8388607...8388607        low byte first
    # S3R       signed 3 byte int reverse    -8388607...8388607        high byte first
    "S3N": types.IntType(-8388607, 8388607),
    "S3R": types.IntType(-8388607, 8388607),
    # ULG       unsigned integer              0...4294967294           low byte first
    # ULR       unsigned integer reverse      0...4294967294           high byte first
    "ULG": types.IntType(0, 4294967294),
    "ULR": types.IntType(0, 4294967294),
    # SLG       signed integer               -2147483647...2147483647  low byte first
    # SLR       signed integer reverse       -2147483647...2147483647  high byte first
    "SLG": types.IntType(-2147483647, 2147483647),
    "SLR": types.IntType(-2147483647, 2147483647),
}


def decode_type(typecode, divider=None):
    """
    Get :any:`Type` instance for `typecode`.

    Args:
        typecode: Type code according to typecodes_

    Keyword Args:
        divider (int): Additional divider.

    .. _typecodes: https://github.com/john30/ebusd/wiki/4.3.-Builtin-data-types

    >>> decode_type("STR:*")
    StrType()
    >>> decode_type("STR:9")
    StrType(length=9)
    >>> decode_type("HEX:*")
    HexType()
    >>> decode_type("HEX:9")
    HexType(length=9)
    >>> decode_type("BI0:0")
    BoolType()
    >>> decode_type("BI0:7")
    BoolType()
    >>> decode_type("BDA")
    DateType()
    >>> decode_type("BDA:3")
    DateType()
    >>> decode_type("HDA")
    DateType()
    >>> decode_type("HDA:3")
    DateType()
    >>> decode_type("DAY")
    DateType()
    >>> decode_type("BTI")
    TimeType()
    >>> decode_type("HTI")
    TimeType()
    >>> decode_type("VTI")
    TimeType()
    >>> decode_type("BTM")
    HourMinuteType()
    >>> decode_type("HTM")
    HourMinuteType()
    >>> decode_type("VTM")
    HourMinuteType()
    >>> decode_type("MIN")
    HourMinuteType()
    >>> decode_type("TTM")
    HourMinuteType(minres=10)
    >>> decode_type("TTH")
    HourMinuteType(minres=30)
    >>> decode_type("TTQ")
    HourMinuteType(minres=15)
    >>> decode_type("BDY")
    WeekdayType()
    >>> decode_type("HDY")
    WeekdayType()
    >>> decode_type("BCD")
    IntType(0, 99)
    >>> decode_type("BCD:2")
    IntType(0, 9999)
    >>> decode_type("BCD:3")
    IntType(0, 999999)
    >>> decode_type("BCD:4")
    IntType(0, 99999999)
    >>> decode_type("PIN")
    PinType()
    >>> decode_type("UCH")
    IntType(0, 254)
    >>> decode_type("SCH")
    IntType(-127, 127)
    >>> decode_type("D1B")
    IntType(-127, 127)
    >>> decode_type("D1C")
    IntType(0, 100, divider=2)
    >>> decode_type("D2B")
    IntType(-127.99, 127.99, divider=256)
    >>> decode_type("D2C")
    IntType(-2047.9, 2047.9, divider=16)
    >>> decode_type("FLT")
    IntType(-32.767, 32.767, divider=1000)
    >>> decode_type("FLR")
    IntType(-32.767, 32.767, divider=1000)
    >>> decode_type("EXP")
    FloatType()
    >>> decode_type("EXR")
    FloatType()
    >>> decode_type("UIN")
    IntType(0, 65534)
    >>> decode_type("UIR")
    IntType(0, 65534)
    >>> decode_type("SIN")
    IntType(-32767, 32767)
    >>> decode_type("SIR")
    IntType(-32767, 32767)
    >>> decode_type("U3N")
    IntType(0, 16777214)
    >>> decode_type("U3R")
    IntType(0, 16777214)
    >>> decode_type("S3N")
    IntType(-8388607, 8388607)
    >>> decode_type("S3R")
    IntType(-8388607, 8388607)
    >>> decode_type("ULG")
    IntType(0, 4294967294)
    >>> decode_type("ULR")
    IntType(0, 4294967294)
    >>> decode_type("SLG")
    IntType(-2147483647, 2147483647)
    >>> decode_type("SLR")
    IntType(-2147483647, 2147483647)
    """
    # create missing types
    if typecode not in TYPEMAP:
        # STR       character string              Hello
        # NTS       character string              Hello
        if typecode.startswith(("STR:", "NTS:")):
            TYPEMAP[typecode] = types.StrType(_get_length(typecode))
        # HEX       hex digit string              hex octet sep by space
        if typecode.startswith("HEX:"):
            TYPEMAP[typecode] = types.HexType(_get_length(typecode))
        # BI0     bit 0                         0...1
        mat = _RE_BIT.match(typecode)
        if mat:
            TYPEMAP[typecode] = types.BoolType()
    # get type
    type_ = TYPEMAP[typecode]
    # divider
    if divider:
        type_ = type_.with_divider(divider)
    return type_


def _get_length(typecode):
    length = typecode.split(":")[1]
    if length != "*":
        return int(length)
    else:
        return None



"""
Circuit Information.

Every EBUS Device has a common set of meta information:

* circuit: The unique circuit name within the bus infrastructure. Duplicates are separated be a suffix `.<idx>`.
* manufacturer: The circuit manufacturer
* model: The circuit model
* swversion: Software Version
* hwversion: Hardware Version
* address: EBUS address
"""
import collections

CircuitInfo = collections.namedtuple("CircuitInfo", "circuit manufacturer model swversion hwversion address")



"""Command Line. Interface."""
import argparse
import asyncio
import inspect
import logging
import sys

from .. import __version__
from . import cmd, dummyserver, info, listen, ls, observe, read, state, write


def argvhandler(argv):
    """Command Line Interface."""
    parser = argparse.ArgumentParser(prog="ebustool")

    parser.add_argument("--host", "-H", default="127.0.0.1", help="EBUSD address. Default is '127.0.0.1'.")
    parser.add_argument("--port", "-P", default=8888, type=int, help="EBUSD port. Default is 8888.")
    parser.add_argument("--timeout", "-T", default=10, type=int, help="EBUSD connection timeout. Default is 10.")

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument("--debug", action="store_true", default=False)
    parser.set_defaults(main=lambda args: _print_help(parser))

    # Sub Commands
    subparsers = parser.add_subparsers(help="Sub Commands")
    cmd.parse_args(subparsers)
    listen.parse_args(subparsers)
    ls.parse_args(subparsers)
    observe.parse_args(subparsers)
    read.parse_args(subparsers)
    state.parse_args(subparsers)
    write.parse_args(subparsers)
    info.parse_args(subparsers)
    dummyserver.parse_args(subparsers)

    args = parser.parse_args(argv)
    loglevel = logging.DEBUG if args.debug else logging.WARN
    logging.basicConfig(format="%(levelname)10s %(name)20s %(message)s", level=loglevel)
    if inspect.iscoroutinefunction(args.main):
        asyncio.run(args.main(args))
    else:
        args.main(args)


def _print_help(parser):
    parser.print_help()
    sys.exit(2)


def main():  # pragma: no cover
    """Command Line Hookup."""
    args = sys.argv[1:]
    if "--debug" in args:
        argvhandler(args)
    else:
        try:
            argvhandler(args)
        except Exception as exc:  # pylint: disable=W0703
            print("ERROR: %s" % exc)
            sys.exit(1)



"""
Virtual Field Definitions.

Fields which are assembled/calcuated from one or more other fields.
"""
import datetime

from .const import NA
from .msgdef import VirtFieldDef
from .types import DateTimeType, DateType, HourMinuteType, TimeType


def iter_virtfielddefs(fielddefs):
    """Iterate over Generic Field Definitions."""
    typeclss = [fielddef.type_.__class__ for fielddef in fielddefs]
    names = [fielddef.name for fielddef in fielddefs]
    if DateType in typeclss and TimeType in typeclss:
        # date and time need to be next to each other
        # Limitation: Just the first pair is found, which should be sufficient
        didx = typeclss.index(DateType)
        tidx = typeclss.index(TimeType)
        if abs(didx - tidx) == 1:
            if "dcfstate" in names:
                sidx = names.index("dcfstate")
                yield VirtFieldDef(
                    f"{names[didx]}+{names[tidx]}+dcfstate",
                    DateTimeType(),
                    lambda fields: _merge_date_time(fields[didx].value, fields[tidx].value, fields[sidx].value),
                )
            else:
                yield VirtFieldDef(
                    f"{names[didx]}+{names[tidx]}",
                    DateTimeType(),
                    lambda fields: _merge_date_time(fields[didx].value, fields[tidx].value),
                )
    if DateType in typeclss and HourMinuteType in typeclss:
        # date and time need to be next to each other
        # Limitation: Just the first pair is found, which should be sufficient
        didx = typeclss.index(DateType)
        tidx = typeclss.index(HourMinuteType)
        if abs(didx - tidx) == 1:
            yield VirtFieldDef(
                f"{names[didx]}+{names[tidx]}",
                DateTimeType(),
                lambda fields: _merge_date_time(fields[didx].value, fields[tidx].value),
            )
    if len(fielddefs) > 1 and names[-1] == "sensor":
        valuedef = fielddefs[0]
        sensordef = fielddefs[-1]
        yield VirtFieldDef(
            f"{valuedef.name}+{sensordef.name}",
            valuedef.type_,
            lambda fields: _merge_sensor_status(fields[valuedef.idx].value, fields[sensordef.idx].value),
            unit=valuedef.unit,
        )


def _merge_date_time(date, time, state=None):
    if date not in (None, NA) and time not in (None, NA):
        if state in (None, "valid", "ok"):
            return datetime.datetime(date.year, date.month, date.day, time.hour, time.minute, time.second)
        else:
            return state
    return NA


def _merge_sensor_status(value, sensor):
    if NA in (value, sensor):
        return NA
    elif sensor == "ok":
        return value
    else:
        return sensor



"""Message Defintions."""
import collections
import copy

from anytree import NodeMixin

from .const import AUTO
from .util import repr_

_MsgDef = collections.namedtuple("_MsgDef", "circuit name read prio write update setprio")


class MsgDef(_MsgDef, NodeMixin):

    """
    Message Definition.

    A message is defined by `circuit` and `name` and has a fields

    Args:
        circuit (str): Circuit Name
        name (str): Message Name
        children (tuple): Field definitions. :any:`FieldDef` and :any:`VirtFieldDef`.

    Keyword Args:
        read (bool): Message intend to be read
        prio (int): Message Polling Priority on bus.
        write (bool): Message intend to be written
        updated (bool): Message intent to be seen automatically on every value change
        setprio: Message polling priority to be set. Integer `1-9 or `A` for automatic.

    >>> from pyebus import MsgDef, types
    >>> m = MsgDef('circuit', 'name', [
    ...     FieldDef(0, 'temp', types.IntType(-127, 128)),
    ...     FieldDef(1, 'name', types.StrType(10)),
    ...     VirtFieldDef('virt', types.StrType(), lambda msg: f'{msg.name}: {msg.temp}'),
    ... ], read=True)
    >>> m  # doctest: +ELLIPSIS
    MsgDef('circuit', 'name', (FieldDef(0, 'temp', IntType(... VirtFieldDef('virt', StrType())), read=True)
    >>> m.ident
    'circuit/name'
    >>> m.children
    (FieldDef(0, 'temp', IntType(-127, 128)), FieldDef(1, 'name', StrType(length=10)), VirtFieldDef('virt', StrType()))
    >>> m.fields
    (FieldDef(0, 'temp', IntType(-127, 128)), FieldDef(1, 'name', StrType(length=10)))
    >>> m.virtfields
    (VirtFieldDef('virt', StrType()),)
    >>> m.access
    'r----'

    :any:`MsgDef`, :any:`FieldDef` and :any:`VirtFieldDef` form a tree structure

    >>> from anytree import RenderTree
    >>> print(RenderTree(m))  # doctest: +ELLIPSIS
    MsgDef('circuit', 'name', (FieldDef(0, 'temp', IntType(-127, 128)),..., StrType())), read=True)
    ├── FieldDef(0, 'temp', IntType(-127, 128))
    ├── FieldDef(1, 'name', StrType(length=10))
    └── VirtFieldDef('virt', StrType())

    Similar object can be easily created by:

    >>> m.replace(children=m.fields[1:2])
    MsgDef('circuit', 'name', (FieldDef(1, 'name', StrType(length=10)),), read=True)
    """

    __slots__ = tuple()

    def __new__(cls, circuit, name, children, read=False, prio=None, write=False, update=False, setprio=None):
        if not read:
            prio = None
        msgdef = _MsgDef.__new__(cls, circuit, name, read, prio, write, update, setprio)
        if children:
            msgdef.children = children
        return msgdef

    def __repr__(self):
        args = (self.circuit, self.name, self.children)
        kwargs = [
            ("read", self.read, False),
            ("prio", self.prio, None),
            ("write", self.write, False),
            ("update", self.update, False),
            ("setprio", self.setprio, None),
        ]
        return repr_(self, args, kwargs)

    def __hash__(self):
        return hash(
            (self.circuit, self.name, self.children, self.read, self.prio, self.write, self.update, self.setprio)
        )

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return (
                self.circuit,
                self.name,
                self.children,
                self.read,
                self.prio,
                self.write,
                self.update,
                self.setprio,
            ) == (
                other.circuit,
                other.name,
                other.children,
                other.read,
                other.prio,
                other.write,
                other.update,
                other.setprio,
            )
        else:
            return NotImplemented

    @property
    def fields(self):
        """Fields."""
        return tuple(child for child in self.children if isinstance(child, FieldDef))

    @property
    def virtfields(self):
        """Calulated Fields."""
        return tuple(child for child in self.children if isinstance(child, VirtFieldDef))

    @property
    def ident(self):
        """Identifier."""
        return f"{self.circuit}/{self.name}"

    @property
    def access(self):
        """Message Access."""
        read = "r" if self.read else "-"
        write = "w" if self.write else "-"
        update = "u" if self.update else "-"
        setprio = str(self.setprio) if self.setprio else "-"
        prio = str(self.prio) if self.prio else "-"
        return "".join((read, write, update, setprio, prio))

    def join(self, msgdef):
        """Return Joined Message Definition."""
        if (self.circuit, self.name, self.children) == (msgdef.circuit, msgdef.name, msgdef.children):
            return self.replace(
                read=self.read or msgdef.read,
                prio=self.prio or msgdef.prio,
                write=self.write or msgdef.write,
                update=self.update or msgdef.update,
                setprio=self.setprio or msgdef.setprio,
            )
        else:
            return None

    def replace(self, **kwargs):
        """Create copy with updated attributes."""
        attrs = self._asdict()
        attrs["children"] = self.children
        attrs.update(kwargs)
        # Take a copy of all children, to ensure proper tree relation
        attrs["children"] = tuple(copy.copy(fielddef) for fielddef in attrs["children"])
        return MsgDef(**attrs)


_FieldDef = collections.namedtuple("_FieldDef", "idx name type_ unit comment")


class AbstractFieldDef(_FieldDef, NodeMixin):

    """
    Abstract Field Definition.

    Args:
        idx (str): Index within Message
        name (str): Unique name (as `name` may be used multiple times by ebus)
        type_ (Type): Type

    Keyword Args:
        unit (str): Unit of the field value
        comment (str): Comment.
    """

    __slots__ = tuple()

    def __new__(cls, idx, name, type_, unit=None, comment=None):
        return _FieldDef.__new__(cls, idx, name, type_, unit or None, comment or None)

    def __repr__(self):
        args = (self.idx, self.name, self.type_)
        kwargs = [
            ("unit", self.unit, None),
            ("comment", self.comment, None),
        ]
        return repr_(self, args, kwargs)

    def _pre_detach(self, parent):
        # it is forbidden to remove fields from their message - create new one
        assert False, f"{self!r} is already used by {parent!r}"  # pragma: no cover

    @property
    def msgdef(self):
        """Message Definition."""
        return self.parent

    @property
    def ident(self):
        """Identifier."""
        return f"{self.msgdef.ident}/{self.name}" if self.msgdef else None

    def __copy__(self):
        return FieldDef(idx=self.idx, name=self.name, type_=self.type_, unit=self.unit)


class FieldDef(AbstractFieldDef):

    """
    Field Definition.

    Args:
        idx (str): Index within Message
        name (str): Unique name (as `name` may be used multiple times by ebus)
        type_ (Type): Type

    Keywords Args:
        unit (str): Unit of the field value
        comment (str): Comment.
    """


class VirtFieldDef(AbstractFieldDef):
    """
    Virtual Field Definition.

    Args:
        name (str): Unique name (as `name` may be used multiple times by ebus)
        type_ (Type): Type
        func: Function to create value.

    Keywords Args:
        unit (str): Unit of the field value
        comment (str): Comment.
    """

    def __new__(cls, name, type_, func, unit=None, comment=None):
        obj = AbstractFieldDef.__new__(cls, None, name, type_, unit or None, comment or None)
        obj.func = func
        return obj

    def __repr__(self):
        return repr_(self, (self.name, self.type_))

    def __copy__(self):
        # pylint: disable=E1101
        return VirtFieldDef(name=self.name, type_=self.type_, func=self.func, unit=self.unit, comment=self.comment)


def resolve_prio(msgdef, setprio=AUTO):
    """
    Resolve priority specification.

    Integer values are just passed through.

    >>> resolve_prio(MsgDef('circuit', 'name', [], read=True), 1)
    1
    >>> resolve_prio(MsgDef('circuit', 'name', [], read=True), 9)
    9

    The AUTO option, sets read-only messages to prio 1.

    >>> resolve_prio(MsgDef('circuit', 'name', [], read=True), AUTO)
    1

    All other readable messages get prio 2

    >>> resolve_prio(MsgDef('circuit', 'name', [], read=True, write=True), AUTO)
    2
    >>> resolve_prio(MsgDef('circuit', 'name', [], read=True, update=True), AUTO)
    2

    Non-readable messages have no prio.

    >>> resolve_prio(MsgDef('circuit', 'name', [], write=True), AUTO)
    """
    if msgdef.read:
        if setprio == AUTO:
            # Prio is only available on read messages.
            if msgdef.update or msgdef.write:
                # we just want to ensure that we do not miss any value
                setprio = 2
            else:
                setprio = 1
        return setprio
    else:
        return None



"""Message Prioritizer."""
import collections
import logging
from datetime import datetime, timedelta

from .exceptions import UnknownMsgError
from .msgdef import resolve_prio

_LOGGER = logging.getLogger(__name__)


class Prioritizer:

    """
    Prioritizer which creates recommended message polling priority according to message change interval.

    Prio 1 is highest.
    The number of priorities is defined by the number of thresholds + 1.

    Args:
        msgdefs (MsgDefs): Message Definitions
        thresholds: Iterable with seconds or :any:`timedelta` between priority levels.
    Keyword Args:
        intervals: Number of intervals the message update rate need to deviate in a row to get a new prio.
    """

    # pylint: disable=R0902
    def __init__(self, msgdefs, thresholds, intervals=3):
        self._msgdefs = msgdefs
        self._thresholds = tuple(_cast_threshold(threshold) for threshold in thresholds)
        self.intervals = intervals
        self.clear()

    def clear(self):
        """Clear."""
        self._msgtimestamps = {}
        self._msgprios = {}
        self._upfilter = collections.defaultdict(lambda: 0)
        self._dnfilter = collections.defaultdict(lambda: 0)
        self._msgpriochanges = {}

    @property
    def thresholds(self):
        """Tuple of timedelta objects with define the time thresholds for each priority."""
        return self._thresholds

    @property
    def maxprio(self):
        """
        Maximum Prio.

        The number of thresholds + 1.
        """
        return len(self._thresholds) + 1

    def set_prio(self, msgdef, prio):
        """Set Message Priority."""
        if msgdef.read:
            self._msgprios[msgdef.ident] = max(min(prio, self.maxprio), 0)
            self._clear_filter(msgdef.ident)

    def get_prio(self, msgdef):
        """Get Message Priority."""
        if msgdef.read:
            return self._get_prio(msgdef)
        else:
            return None

    def _get_prio(self, msgdef):
        try:
            return self._msgprios[msgdef.ident]
        except KeyError:
            return resolve_prio(msgdef)

    def iter_priochanges(self):
        """
        Iterate over done priority changes.

        Yields
            (MsgDef)
        """
        msgpriochanges = self._msgpriochanges
        while msgpriochanges:
            yield msgpriochanges.popitem()[1]

    def notify(self, msg):
        """
        Notify about received message.

        Raise:
            UnknownMsgError
        """
        msgdef = msg.msgdef
        if not self._msgdefs.get_ident(msgdef.ident):
            raise UnknownMsgError(f"circuit={msgdef.circuit}, name={msgdef.name}")
        if msg.valid and msgdef.read:
            msgtimestamps = self._msgtimestamps
            ident = msg.ident
            values = tuple(field.value for field in msg.fields)
            timestamp = datetime.now()
            if ident not in msgtimestamps:
                # unknown, store first entry
                msgtimestamps[ident] = (timestamp, values)
                prio = msgdef.setprio or msgdef.prio or self._get_prio(msg.msgdef)
                self._msgprios[ident] = prio
                _LOGGER.debug("Init   %s, prio=%d", ident, prio)
            else:
                (lasttimestamp, lastvalues) = msgtimestamps[ident]
                age = timestamp - lasttimestamp
                prio = self._get_prio(msg.msgdef)
                is_const = values == lastvalues
                older = prio < self.maxprio and age > self._thresholds[prio - 1]
                newer = prio > 1 and age < self._thresholds[min(prio, self.maxprio) - 2]
                _LOGGER.debug(
                    "Notify %s: age=%r prio=%r is_const=%r older=%r newer=%r", ident, age, prio, is_const, older, newer
                )
                # check for increment
                #   - reset filter, if not const
                #   - increment, if constant value is older than actual threshold for number of intervals
                if not is_const or older:
                    if self._filter(self._upfilter, ident, is_const or older):
                        _LOGGER.debug("Inc!   %s", ident)
                        self._set_prio(ident, prio + 1)
                    msgtimestamps[ident] = (timestamp, values)
                # check for decrement
                #   - reset filter, if constant value
                if self._filter(self._dnfilter, ident, not is_const and newer):
                    _LOGGER.debug("Dec!   %s", ident)
                    self._set_prio(ident, prio - 1)

    def _set_prio(self, ident, prio):
        self._msgprios[ident] = prio
        msgdef = self._msgdefs.get_ident(ident)
        self._msgpriochanges[ident] = msgdef.replace(setprio=prio)

    def _clear_filter(self, ident):
        self._upfilter[ident] = 0
        self._dnfilter[ident] = 0

    def _filter(self, cnts, ident, cond):
        cnt = cnts[ident]
        if cond and cnt < self.intervals:
            cnt += 1
            cond = False
        else:
            cnt = 0
        cnts[ident] = cnt
        return cond


def _cast_threshold(threshold):
    if not isinstance(threshold, timedelta):
        threshold = timedelta(seconds=threshold)
    return threshold



"""Constants."""
from datetime import timedelta


class _NotAvailable:

    """Not Available."""

    def __str__(self):
        return "NA"

    def __repr__(self):
        return "NA"


OK = "ok"
AUTO = "A"
NA = _NotAvailable()

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 8888
DEFAULT_TIMEOUT = 10
DEFAULT_PRIOTHRESHOLDS = [timedelta(hours=1), timedelta(hours=4), timedelta(days=1)]



"""Write Command."""
from .common import add_msgdef_args, async_load_msgdefs, create_ebus, disable_stdout_buffering


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("write", help="Write value to the bus")
    add_msgdef_args(parser)
    parser.add_argument("field", help="Field (i.e. 'ui/OutsideTemp/temp')")
    parser.add_argument("value", help="Value to apply (i.e. '5'). 'NONE' is reserved for no value.")
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    await async_load_msgdefs(ebus, args)
    value = args.value if args.value != "NONE" else None
    for msgdef in ebus.msgdefs.resolve(args.field.split(";"), filter_=lambda msgdef: msgdef.write):
        await ebus.async_write(msgdef, value)



"""EBUS Messages And Their Fields."""
import collections

from .const import NA
from .util import repr_


class Msg(collections.namedtuple("_Msg", "msgdef fields")):

    """
    Message with Fields.

    The message knows its definition and its fields.
    """

    __slots__ = tuple()

    valid = True

    def __repr__(self):
        args = (self.msgdef.ident, self.fields)
        return repr_(self, args)

    @property
    def ident(self):
        """Identifier."""
        return self.msgdef.ident

    @property
    def values(self):
        """Values."""
        return tuple(field.value for field in self.fields)


class Field(collections.namedtuple("_Field", "fielddef value")):

    """
    Field with Data.

    The field knows its defintion and have one value.
    """

    __slots__ = tuple()

    def __repr__(self):
        args = (self.fielddef.name, self.value)
        return repr_(self, args)

    @property
    def ident(self):
        """Identifier."""
        return self.fielddef.ident

    @property
    def unitvalue(self):
        """Unitized Value."""
        value = self.value
        if value is not None and value is not NA and not isinstance(value, str) and self.fielddef.unit:
            return f"{value} {self.fielddef.unit}"
        else:
            return value


def filter_msg(msg=None, msgdefs=None):
    """Strip Down Message according to `msgdefs`."""
    if msg is not None:
        ident = msg.msgdef.ident
        if msgdefs is not None:
            for msgdef in msgdefs:
                if ident == msgdef.ident:
                    if msg.msgdef == msgdef or not msg.valid:
                        return msg
                    else:
                        fields = tuple(
                            Field(field.fielddef, field.value)
                            for field in msg.fields
                            if field.fielddef in msgdef.fields
                        )
                        return Msg(msgdef, fields)
        else:
            return msg
    return None


class BrokenMsg:

    """
    Broken Message.

    A broken message failed during decoding of EBUSD or the :any:`MsgDecoder`.
    """

    valid = False

    def __init__(self, msgdef, error):
        """
        Broken Message.

        .. note: In a boolean context this instance evaluates to `False`.
        """
        self.msgdef = msgdef
        self.error = error

    def __repr__(self):
        return repr_(self, (self.msgdef.ident, self.error))

    @property
    def values(self):
        """Values."""
        return tuple()



"""Dummy Server emulating EBUSD instance."""
import collections
import logging
import re

from .dummydata import DummyData

_LOGGER = logging.getLogger()
_RE_READ = re.compile(r"read -c ([A-Za-z0-9\.]+)( -m \d+)?( -p (\d+))? ([A-Za-z0-9\.]+)")


class Dummy:

    """
    Partial Dummy Server emulating EBUSD instance, just for testing.

    Keyword Args:
        port (int): Port. Default is 8888.
    """

    def __init__(self, dummydata=None):
        self.dummydata = dummydata or DummyData()
        self.data = collections.defaultdict(lambda: 0)
        self.prios = {}
        self.notwriteable = []

    def respond(self, request):
        """Iterate over response for `request`."""
        if request == "quit":  # pragma: no cover
            yield "connection closed"
            yield ""
            raise Quit()
        if request == "stop":
            yield "stopping"
            yield ""
            raise Stop()

        resp = {
            "state": [self.dummydata.state],
            "info": self.dummydata.info,
            "find -a -F type,circuit,name,fields": self.dummydata.finddef,
            "find -d": self.dummydata.finddata,
            "dummyshutdown": ["ERR: shutdown"],
            "dummytrailing": ["ok", "ok2"],
        }.get(request, None)
        m_read = _RE_READ.match(request)

        if resp:
            yield from resp
        elif m_read:
            circuit, _, _, prio, name = m_read.groups()
            if prio is not None:
                self.prios[(circuit, name)] = prio
            yield str(self.data[(circuit, name)])
        elif request.startswith("write -c "):
            (circuit, name, value) = request[len("write -c ") :].split(" ")
            if (circuit, name) not in self.notwriteable:
                self.data[(circuit, name)] = value
                yield "done"
            else:
                yield "ERR: not writable"
        elif request == "listen":
            if self.dummydata.listen:
                yield "listen started"
                yield ""
                yield from self.dummydata.listen
            else:
                yield "listen broken"
        else:
            yield "ERR: command not found"
            _LOGGER.error("%r: unknown request=%r", self, request)
        yield ""


class Quit(RuntimeError):

    """Quit."""


class Stop(RuntimeError):

    """Stop."""



"""Write Command."""
from pyebus import DummyServer


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("dummyserver", help="Run Dummy Server, which behaves like EBUSD")
    parser.add_argument("--port", type=int, help="Listen on Port", default=8888)
    parser.set_defaults(main=_main)


async def _main(args):
    server = DummyServer(port=args.port)
    await server.async_start()
    await server.async_wait_closed()



"""Listen Command."""

from .common import (
    add_msgdef_args,
    add_patterns_arg,
    async_load_msgdefs,
    create_ebus,
    disable_stdout_buffering,
    print_msg,
)


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("listen", help="Listen on the bus, decode messages and print")
    add_msgdef_args(parser)
    add_patterns_arg(parser, opt=True)
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    await async_load_msgdefs(ebus, args)
    msgdefs = ebus.msgdefs.resolve(args.patterns)
    print(f"Listening to {msgdefs.summary()}")
    if msgdefs:
        async for msg in ebus.async_listen(msgdefs=msgdefs):
            print_msg(msg)



"""EBUS Connection Handling."""
import collections
import logging

from .connection import CommandError, Connection, Shutdown
from .const import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT
from .dummy import Dummy

_LOGGER = logging.getLogger(__name__)


class DummyConnection(Connection, Dummy):
    """
    Dummy EBUS Connection.

    Keyword Args:
        host (str): Hostname or IP
        port (int): Port
        autoconnect (bool): Automatically connect and re-connect
        timeout (int): Connection Timeout
        dummydata (DummyData): storage for responses
    """

    def __init__(
        self, host=DEFAULT_HOST, port=DEFAULT_PORT, autoconnect=False, timeout=DEFAULT_TIMEOUT, dummydata=None
    ):
        Connection.__init__(self, host=host, port=port, autoconnect=autoconnect, timeout=timeout)
        Dummy.__init__(self, dummydata=dummydata)
        self.__connected = False
        self.__respbuffer = collections.deque()

    async def async_connect(self):
        """
        Establish connection (required before first communication).

        Raises:
            ConnectionRefusedError: If connection cannot be established
        """
        _LOGGER.debug("connect()")
        self.__connected = True

    async def async_disconnect(self):
        """Disconnect if not already done."""
        _LOGGER.debug("disconnect()")
        self.__connected = False

    def is_connected(self):
        """
        Return `True` if connection is established.

        This does not check if the connection is still usable.

        Returns:
            bool
        """
        return self.__connected

    async def _async_write(self, message):
        assert not self.__respbuffer, "Response Buffer not empty"
        self.__respbuffer.extend(self.respond(message))

    async def _async_readline(self, check=False):
        line = self.__respbuffer.popleft()
        _LOGGER.debug(f"_readline() = {line!r}")
        if line == "ERR: shutdown":
            raise Shutdown()
        if check and line.startswith("ERR: "):
            raise CommandError(line.lstrip("ERR: "))
        return line

    async def _async_ensure_connection(self):
        if not self.__connected:
            if self._autoconnect:
                await self.async_connect()
            else:
                raise ConnectionError("Not connected")



"""Not Available."""


class NotAvailable:

    """Not Available Singleton."""

    __instance = None

    def __new__(cls):
        if NotAvailable.__instance is None:
            NotAvailable.__instance = object.__new__(cls)
        return NotAvailable.__instance

    def __str__(self):
        return "Not Available"

    def __repr__(self):
        return "NA"

    def __bool__(self):
        return False


NA = NotAvailable()



"""EBUS Connection Handling."""
import asyncio
import logging

from .const import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT
from .exceptions import CommandError, Shutdown
from .util import repr_

_LOGGER = logging.getLogger(__name__)


class Connection:
    """
    EBUS Connection.

    Keyword Args:
        host (str): Hostname or IP
        port (int): Port
        autoconnect (bool): Automatically connect and re-connect
        timeout (int): Connection Timeout
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT, autoconnect=False, timeout=DEFAULT_TIMEOUT):
        self._host = host
        self._port = port
        self._autoconnect = autoconnect
        self._timeout = timeout
        self._reader, self._writer = None, None

    def __repr__(self):
        return repr_(
            self,
            kwargs=(
                ("host", self.host, DEFAULT_HOST),
                ("port", self.port, DEFAULT_PORT),
                ("autoconnect", self.autoconnect, False),
                ("timeout", self.timeout, DEFAULT_TIMEOUT),
            ),
        )

    @property
    def host(self):
        """Host."""
        return self._host

    @property
    def port(self):
        """Port."""
        return self._port

    @property
    def autoconnect(self):
        """Automatically connect and re-connect."""
        return self._autoconnect

    @property
    def timeout(self):
        """Connection Timeout."""
        return self._timeout

    async def async_connect(self):
        """
        Establish connection (required before first communication).

        Raises:
            ConnectionRefusedError: If connection cannot be established
        """
        _LOGGER.debug("connect()")
        self._reader, self._writer = await self._async_timedout(asyncio.open_connection(self._host, self._port))

    async def async_disconnect(self):
        """Disconnect if not already done."""
        _LOGGER.debug("disconnect()")
        if self._writer:
            try:
                await self._async_write("quit")
                self._writer.close()
                await self._writer.wait_closed()
            except BrokenPipeError:  # pragma: no cover
                pass
            finally:
                self._reader, self._writer = None, None

    def is_connected(self):
        """
        Return `True` if connection is established.

        This does not check if the connection is still usable.

        Returns:
            bool
        """
        return self._writer is not None and not self._writer.is_closing()

    async def async_write(self, message):
        """
        Send TCP `message` to EBUSD.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: If not connected (`autoconnect==False`)
        """
        _LOGGER.debug(f"write({message!r})")
        await self._async_ensure_connection()
        await self._async_write(message)

    async def async_request(self, cmd, *args, **kwargs):
        """
        Assemble request starting with `cmd` and position `args` and keywords `kwargs'.

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: If not connected (`autoconnect==False`)
        """
        parts = [cmd]
        parts += [f"-{option} {value}" for option, value in kwargs.items() if value is not None]
        parts += [str(arg) for arg in args]
        message = " ".join(parts)
        _LOGGER.debug(f"request({message!r})")
        await self._async_ensure_connection()
        await self._async_write(message)

    async def async_read(self, infinite=False, check=True):
        """
        Receive lines until an empty one (`infinite==False`) or infinitly (`infinite==True`).

        Yields:
            str: line read

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: If not connected (`autoconnect==False`)
            CommandError: If command failed (`check==True`)
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.debug(f"read(infinite={infinite!r}, check={check!r})")
        await self._async_ensure_connection()
        while True:
            line = await self._async_readline(check=check)
            if not line and not infinite:
                break
            yield line

    async def async_readresp(self, check=True):
        """
        Receive command response.

        Read one line as command response and one empty line.

        Returns:
            str: response

        Raises:
            ConnectionRefusedError: If connection cannot be established
            ConnectionError: If not connected (`autoconnect==False`)
            CommandError: If command failed (`check==True`) or trailing data.
            Shutdown: On EBUSD shutdown.
        """
        _LOGGER.debug(f"readresp(check={check!r})")
        await self._async_ensure_connection()
        line = await self._async_readline(check=check)
        empty = await self._async_readline()
        if empty:
            raise CommandError(f"Trailing data {empty}")
        return line

    async def _async_write(self, message):
        self._writer.write(f"{message}\n".encode())
        await self._async_timedout(self._writer.drain())

    async def _async_readline(self, check=False):
        line = await self._reader.readline()
        line = line.decode("utf-8").rstrip()
        _LOGGER.debug(f"_readline() = {line!r}")
        if line == "ERR: shutdown":
            raise Shutdown()
        if check and line.startswith("ERR: "):
            raise CommandError(line.lstrip("ERR: "))
        return line

    async def _async_ensure_connection(self):
        if not self._writer or self._writer.is_closing():
            if self._autoconnect:
                await self.async_connect()
            else:
                raise ConnectionError("Not connected")

    async def _async_timedout(self, task):
        if self._timeout:
            try:
                result = await asyncio.wait_for(task, timeout=self._timeout)
            except asyncio.TimeoutError as timeout:
                raise ConnectionError(f"{self.host}:{self.port} timeout") from timeout
        else:
            result = await task
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyebus', 'pyebus.cli']

package_data = \
{'': ['*']}

install_requires = \
['anytree>=2.8.0,<3.0.0']

entry_points = \
{'console_scripts': ['ebt = pyebus.cli:main']}

setup_kwargs = {
    'name': 'pyebus',
    'version': '1.2.4',
    'description': 'Pythonic Interface to EBUS Daemon (ebusd)',
    'long_description': '.. image:: https://badge.fury.io/py/pyebus.svg\n    :target: https://badge.fury.io/py/pyebus\n\n.. image:: https://img.shields.io/pypi/dm/pyebus.svg?label=pypi%20downloads\n   :target: https://pypi.python.org/pypi/pyebus\n\n.. image:: https://travis-ci.org/c0fec0de/pyebus.svg?branch=master\n    :target: https://travis-ci.org/c0fec0de/pyebus\n\n.. image:: https://readthedocs.org/projects/pyebus/badge/?version=latest\n    :target: https://pyebus.readthedocs.io/en/latest/?badge=latest\n\n.. image:: https://codeclimate.com/github/c0fec0de/pyebus.png\n    :target: https://codeclimate.com/github/c0fec0de/pyebus\n\n.. image:: https://img.shields.io/pypi/pyversions/pyebus.svg\n   :target: https://pypi.python.org/pypi/pyebus\n\n.. image:: https://img.shields.io/badge/code%20style-pep8-brightgreen.svg\n   :target: https://www.python.org/dev/peps/pep-0008/\n\n.. image:: https://img.shields.io/badge/code%20style-pep257-brightgreen.svg\n   :target: https://www.python.org/dev/peps/pep-0257/\n\nPythonic interface to EBUSD_.\n\n\nInstallation\n============\n\nTo install the `pyebus` module run::\n\n    pip install pyebus\n\nIf you do not have write-permissions to the python installation, try::\n\n    pip install pyebus --user\n\nCommand-Line-Interface\n======================\n\nUsage::\n\n\tusage: ebustool [-h] [--host HOST] [--port PORT] [--timeout TIMEOUT] [--version] [--debug] {cmd,listen,ls,observe,read,state,write,info} ...\n\n\tpositional arguments:\n\t  {cmd,listen,ls,observe,read,state,write,info}\n\t                        Sub Commands\n\t    cmd                 Issue TCP Command on EBUSD. See https://github.com/john30/ebusd/wiki/3.1.-TCP-client-commands for reference.\n\t    listen              Listen on the bus, decode messages and print\n\t    ls                  List all messages\n\t    observe             Read all known messages once and continue listening so that ALL EBUS values are available, decode every message and print.\n\t    read                Read values from the bus, decode and print\n\t    state               Show EBUSD state\n\t    write               Write value to the bus\n\t    info                Show EBUSD meta information\n\n\toptional arguments:\n\t  -h, --help            show this help message and exit\n\t  --host HOST, -H HOST  EBUSD address. Default is \'127.0.0.1\'.\n\t  --port PORT, -P PORT  EBUSD port. Default is 8888.\n\t  --timeout TIMEOUT, -T TIMEOUT\n\t                        EBUSD connection timeout. Default is 10.\n\t  --version             show program\'s version number and exit\n\t  --debug\n\nList all messages and fields\n----------------------------\n\nThe `ls` command lists all messages `CIRCUIT/MESSAGENAME/FIELDNAME     rwuSP INFO`::\n\n\t$ ebt -H 192.168.1.4 ls\n\tmc/DateTime/dcfstate                     r---1 nosignal, ok, sync, valid [DCF Empfängerstatus]\n\tmc/DateTime/btime                        r---1 HOUR:MINUTE:SECOND [Uhrzeit]\n\tmc/DateTime/bdate                        r---1 DAY.MONTH.YEAR [Datum]\n\tmc/DateTime/temp2                        r---1 Float within [-127.99:127.99] with 0.00390625 fraction\n\tmc/DateTime/bdate+btime+dcfstate         r---1 DAY.MONTH.YEAR HOUR:MINUTE:SECOND\n\tmc/FlowTemp/temp                         r---1 Float within [-2047.9:2047.9] with 0.0625 fraction [Temperatur]\n\tmc/FlowTemp/sensor                       r---1 ok, circuit, cutoff\n\tmc/FlowTemp/temp+sensor                  r---1 Float within [-2047.9:2047.9] with 0.0625 fraction\n\tmc/FlowTempDesired/temp1                 r---1 Float within [0:100] with 0.5 fraction\n\tmc/FlowTempMax/temp0                     rw--2 Integer within [0:254]\n\tmc/FlowTempMin/temp0                     rw--2 Integer within [0:254]\n\tmc/OperatingMode/mcmode                  rw--2 disabled, on, off, auto, eco, low\n\tmc/TempDesired/temp1                     rw--2 Float within [0:100] with 0.5 fraction\n\tmc/TempDesiredLow/temp1                  rw--2 Float within [0:100] with 0.5 fraction\n\n`CIRCUIT/MESSAGENAME/FIELDNAME` is a unique message field identifier.\n`CIRCUIT` names the device which contains the information.\n`MESSAGENAME` is the name of the message on the bus.\nEach message consists of fields.\n`FIELDNAME` identifies the specific information within the message.\nEBUSD fieldnames are **NOT** unique. `pyebus` appends a suffix in case of naming collisions.\nField names with a `+` are virtual and just the concatenation of existing fields.\n\nThe access rights have the following meaning:\n\n* `r`: The message is explicitly readable.\n* `w`: The message is explicitly writable.\n* `u`: The message is **NOT** readable, but emitted by the sender on change.\n* `S`: Is the poll priority to be set. **This applies to readable messages only.** Values `1-9` can be used. `A` is a placeholder for automatic. This will choose an appropriate priority.\n* `P`: Actual polling priority.\n\nList some messages and fields\n-----------------------------\n\nThe `ls` command accepts explicit names with wildcards and placeholders. **Case-Insensitive**::\n\n\t$ ebt -H 192.168.1.4 ls "mc/FlowTemp*"\n\tmc/FlowTemp/temp                         r---1 Float within [-2047.9:2047.9] with 0.0625 fraction [Temperatur]\n\tmc/FlowTemp/sensor                       r---1 ok, circuit, cutoff\n\tmc/FlowTemp/temp+sensor                  r---1 Float within [-2047.9:2047.9] with 0.0625 fraction\n\tmc/FlowTempDesired/temp1                 r---1 Float within [0:100] with 0.5 fraction\n\tmc/FlowTempMax/temp0                     rw--2 Integer within [0:254]\n\tmc/FlowTempMin/temp0                     rw--2 Integer within [0:254]\n\n\t$ ebt -H 192.168.1.4 ls "*/*mode*"\n\tmc/OperatingMode/mcmode                  rw--2 disabled, on, off, auto, eco, low\n\nRead messages and fields\n------------------------\n\n`read` behaves identical to `ls` (with or without patterns), but returns the actual value::\n\n\t$ ebt -H 192.168.1.4 read "*/*mode*"\n\tmc/OperatingMode/mcmode                  rw--2 eco                 disabled, on, off, auto, eco, low\n\nNon-readable messages are filtered automatically.\n\nPlease note, EBUS is slow. EBUSD only reads values which are older than 300s or not cached.\n`--ttl` explicitly specifies the maximum age in seconds.\n\nWrite Message Field\n-------------------\n\nEach writable field can be set by::\n\n\t$ ebt -H 192.168.1.4 write mc/OperatingMode/mcmode auto\n\nEBUS Status\n-----------\n\nThe EBUS status can be retrieved by::\n\n\t$ ebt -H 192.168.1.4 state\n\tok\n\nor more detailled::\n\n\t$ ebt -H 192.168.1.4 info\n\tversion                ebusd 21.1.v21.1-12-gccfc025\n\tupdate check           version 3.4 available\n\tsignal                 acquired\n\tsymbol rate            114\n\tmax symbol rate        217\n\tmin arbitration micros 317\n\tmax arbitration micros 4751\n\tmin symbol latency     0\n\tmax symbol latency     10\n\treconnects             0\n\tmasters                7\n\tmessages               1006\n\tconditional            14\n\tpoll                   597\n\tupdate                 10\n\taddress 03             master #11\n\taddress 08             slave #11, scanned "MF=Vaillant;ID=BAI00;SW=0204;HW=9602", loaded "vaillant/bai.0010015600.inc" ([HW=9602]), "vaillant/08.bai.csv"\n\taddress 10             master #2\n\taddress 15             slave #2, scanned "MF=Vaillant;ID=UI   ;SW=0508;HW=6201", loaded "vaillant/15.ui.csv"\n\taddress 17             master #17\n\n\n\nProgramming API\n===============\n\nComplete API-Documentation_\n\nOverview\n--------\n\n`Ebus` represents one connection to a EBUSD instance.\n\n>>> from pyebus import Ebus\n>>> ebus = Ebus(\'127.0.0.1\')\n\nThe instance needs to know the messages definitions handled by the EBUSD instance.\nThe EBUSD message definitions are based on a EBUSD scan algorithm and the EBUSD-configuration_.\n\nWait for the EBUSD scan to be completed:\n\n>>> await ebus.async_wait_scancompleted()\n\nNow, we can load the message definition codes and convert them to message definitions `MsgDef`.\n\n>>> await ebus.async_load_msgdefs()\n\nThe message definitions are stored in a `MsgDefs` instance at\n\n>>> ebus.msgdefs\n\nA single message can be read via:\n\n>>> msgdef = ebus.msgdefs.get(\'circuit\', \'name\')\n>>> print(await ebus.async_read(msgdef))\n\nAll messages can be read via:\n\n>>> for msgdef in ebus.msgdefs:\n>>>   if msgdef.read:\n>>>     print(await ebus.async_read(msgdef))\n\nMessage can be filtered via patterns `CIRCUIT/MESSAGENAME` or `CIRCUIT/MESSAGENAME/FIELDNAME`.\nWildcards (`*`) and Placeholder (`?`) are supported.\nThe following example will read all temperature fields and all message of the circuit named `mc`.\nThe read value must not be older than 1000s.\n\n>>> for msgdef in ebus.msgdefs.resolve([\'*/*/*temp*\', \'mc/*\']):\n>>>   if msgdef.read:\n>>>     print(await ebus.async_read(msgdef, ttl=1000))\n\nWriting is also possible\n\n>>> await ebus.async_write(msgdef, value)\n\nEBUS meta informations are available at:\n\n>>> await ebus.async_get_state()\n>>> await ebus.async_get_info()\n\n.. _EBUSD: https://github.com/john30/ebusd\n.. _EBUSD-Configuration: https://github.com/john30/ebusd-configuration\n.. _API-Documentation: https://pyebus.readthedocs.io/en/latest/api/pyebus.html\n',
    'author': 'c0fec0de',
    'author_email': 'c0fec0de@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/c0fec0de/pyebus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



"""Dummy Data."""


class DummyData:
    """Dummy Data Container."""

    def __init__(self):
        self.info = [
            "version: ebusd 21.1.v21.1-12-gccfc025",
            "update check: version 3.4 available",
            "signal: acquired",
            "symbol rate: 48",
            "max symbol rate: 229",
            "min arbitration micros: 301",
            "max arbitration micros: 3550",
            "min symbol latency: 2",
            "max symbol latency: 15",
            "reconnects: 0",
            "masters: 7",
            "messages: 1006",
            "conditional: 14",
            "poll: 171",
            "update: 10",
            "address 03: master #11",
            'address 08: slave #11, scanned "MF=Vaillant;ID=BAI00;SW=0204;HW=9602", loaded "vaillant/bai.0010015600.inc" ([HW=9602]), "vaillant/08.bai.csv"',  # pylint: disable=C0301
            "address 10: master #2",
            'address 15: slave #2, scanned "MF=Vaillant;ID=UI   ;SW=0508;HW=6201", loaded "vaillant/15.ui.csv"',
            "address 17: master #17",
            'address 1c: slave #17, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/1c.rcc.4.csv"',
            'address 23: slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/23.vr630.cc.csv"',
            'address 25: slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/25.vr630.hwc.csv"',
            'address 26: slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/26.vr630.hc.csv"',
            "address 31: master #8, ebusd",
            "address 36: slave #8, ebusd",
            "address 37: master #18",
            'address 3c: slave #18, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/3c.rcc.5.csv"',
            'address 50: slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/50.vr630.mc.csv"',
            'address 51: slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/51.vr630.mc.3.csv"',
            'address 52: slave, scanned "MF=Vaillant;ID=MC2  ;SW=0500;HW=6301", loaded "vaillant/52.mc2.mc.4.csv"',
            'address 53: slave, scanned "MF=Vaillant;ID=MC2  ;SW=0500;HW=6301", loaded "vaillant/53.mc2.mc.5.csv"',
            "address 70: master #4",
            'address 75: slave #4, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/75.rcc.csv"',
            "address f0: master #5",
            'address f5: slave #5, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/f5.rcc.3.csv"',
        ]
        self.state = "signal acquired, 111 symbols/sec (229 max), 7 masters"

        self.finddef = [
            "r,bai,AntiCondensValue,power,s,UCH,,kW,Anhebung der minimalen Leistung",
            "[SW>=413]r,bai,APCLegioProtection,,s,UCH,,,Legionellenschutz für internen Speicher",
            "r,bai,averageIgnitiontime,,s,UCH,10,s,Mittlere Zündzeit",
            "r,bai,BlockTimeHcMax,minutes0,s,UCH,,min,Max. Brennersperrzeit bei einem Vorlaufsollwert von 20°C",
            "r,bai,BoilerType,,s,UCH,,,Boiler typ of the bmu",
            "r,bai,ChangesDSN,,s,UCH,,,Anzahl der DSN (Gerätekennung) Änderungen",
            "r,bai,CirPump,onoff,s,UCH,0=off;1=on,,Status Zirkulationspumpe (über ein externes Modul ansteuerbar)",
            "r,bai,CodingResistor,,s,HEX:3,,,Therme Identifikation Widerstand",
            "r,bai,CounterStartattempts1,temp0,s,UCH,,°C,Anzahl der erfolglosen Zündversuche (im 1. Versuch)",
            "r,bai,CounterStartattempts2,temp0,s,UCH,,°C,Anzahl der erfolglosen Zündversuche (im 2. Versuch)",
            "r,bai,CounterStartAttempts3,temp0,s,UCH,,°C,Anzahl der erfolglosen Zündversuche (im 3. Versuch)",
            "r,bai,CounterStartAttempts4,temp0,s,UCH,,°C,Anzahl der erfolglosen Zündversuche (im 4. Versuch)",
            "r,bai,dcfState,dcfstate,s,UCH,0=nosignal;1=ok;2=sync;3=valid,,DCF Status",
            "r,bai,DCRoomthermostat,onoff,s,UCH,0=off;1=on,,Wärmeanforderung vom externen Regler (Klemme 3-4)",
            "r,bai,DeactivationsTemplimiter,,s,UCH,,,Anzahl der Abschaltungen durch den Sicherheitstemperaturbgrenzers",
            "r,bai,DeltaFlowReturnMax,temp,s,D2C,,°C,Wartungsdaten",
            "r,bai,DisplayMode,,s,UCH,,,Display mode of the aplliance",
            "r,bai,DSN,,s,UIN,,,DSN: Device Specific number",
            "r,bai,DSNOffset,,s,UCH,,,Gerätekennung (DSN)",
            "r,bai,DSNStart,,s,UIN,,,DSN Startadresse",
            "r,bai,EbusSourceOn,onoff,s,UCH,0=off;1=on,,Aktivierung der eBUS Speisung",
            "r,bai,EbusVoltage,onoff,s,UCH,0=off;1=on,,Rückmeldung der eBUS Spannung",
            "r,bai,ExternalFaultmessage,onoff,s,UCH,0=off;1=on,,Signal für die externe Störmeldeeinrichtung",
            "r,bai,ExternGasvalve,onoff,s,UCH,0=off;1=on,,Externes Magnetventil",
            "r,bai,ExtFlowTempDesiredMin,temp,s,D2C,,°C,minimum out of Kl.7 and eBus flow setpoint",
            "r,bai,ExtStorageModulCon,yesno,s,UCH,0=no;1=yes,,Externes Speichermodul (VR65) angeschlosssen",
            "r,bai,extWP,onoff,s,UCH,0=off;1=on,,Externe Heizungspumpe",
            "r,bai,FanHours,hoursum2,s,UIN,,h,Betriebsstunden des Lüfters",
            "scan",
            "rw,bai,FlowTemp,temp,s,D2C,,°C,Vorlauftemperatur,sensor,s,UCH,0=ok;85=circuit;170=cutoff,,Fühlerstatus",
            "rw,bai,FanPWMSum,,s,UIN,,,Predictive Maintenance data for the fan damage recognition",
            "r,bai,FanPWMTest,,s,UCH,,,Predictive Maintenance data for the fan damage recognition",
        ]
        self.finddata = ["bai FlowTemp = 6.125;ok"]

        self.listen = [
            "bai FlowTemp = 0.125;ok",
            "bai FlowTemp = 1.125;ok",
            "bai FlowTemp = 2.125;ok",
            "bai FlowTemp = ",
            "bai FlowTemp = broken;-",
            "bai FlowTemp2 = broken;-",
            "bai FlowTemp = 3.125;ok",
        ]



"""EBUS Message Decoding."""
import re

from .const import NA
from .exceptions import UnknownMsgError
from .msg import BrokenMsg, Field, Msg


class MsgDecoder:

    """
    Message Decoder.

    Args:
        msgdefs (MsgDefs): Message Definitions

    The message decoder takes a EBUSD data one-line string and creates the corresponding :any:`Msg` instance.
    The decoder needs to know the actual message definitions.
    """

    _re_decode = re.compile(r"([A-z0-9]+(\.[A-z0-9]+)?) ([^\s]*) (= )?(.*)")

    def __init__(self, msgdefs):
        self.msgdefs = msgdefs

    def decode_line(self, line):
        """
        Decode EBUSD data `line` and return :any:`Msg` instance.

        Raises:
            ValueError: if `line` does not match expected format.
            UnknownMsgError: if `line` is not covered by fields.
        """
        match = self._re_decode.match(line)
        if not match:
            raise ValueError(line)
        circuit, _, name, _, valuestr = match.groups()
        msgdef = self.msgdefs.get(circuit, name)
        if not msgdef:
            raise UnknownMsgError(f"circuit={circuit}, name={name}")
        return self.decode_value(msgdef, valuestr.strip())

    def decode_value(self, msgdef, valuestr):
        """
        Decode message `msgdef` value pair string `valuestr`.


        Returns:
            Msg: message with proper data.
            BrokenMsg: Undecodable message.
        """
        if not valuestr.startswith("ERR: "):
            fields = tuple(self._decodefields(msgdef, valuestr.strip()))
            return Msg(msgdef, fields)
        else:
            return BrokenMsg(msgdef, valuestr[len("ERR: ") :].strip())

    @staticmethod
    def _decodefields(msgdef, valuestr):
        if valuestr not in ("no data stored", "nosignal"):
            values = valuestr.split(";")
            fields = []
            for fielddef in msgdef.fields:
                if fielddef.idx is None:
                    continue
                try:
                    value = values[fielddef.idx].strip()
                except IndexError:
                    fieldvalue = NA
                else:
                    try:
                        fieldvalue = fielddef.type_.decode(value)
                    except ValueError:
                        fieldvalue = None
                fields.append(Field(fielddef, fieldvalue))
        else:
            fields = [Field(fielddef, NA) for fielddef in msgdef.fields]
        # virtual fields
        for virtfielddef in msgdef.virtfields:
            virtfieldvalue = virtfielddef.func(fields)
            fields.append(Field(virtfielddef, virtfieldvalue))
        return fields



"""List Command."""

from ..msgdefdecoder import decodetype
from .common import add_msgdef_args, add_patterns_arg, async_load_msgdefs, create_ebus, disable_stdout_buffering


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser("ls", help="List all messages")
    add_msgdef_args(parser)
    add_patterns_arg(parser, opt=True)
    parser.add_argument(
        "--name-only",
        "-n",
        default=False,
        action="store_true",
        help="Just print names.",
    )
    parser.add_argument(
        "--type",
        "-t",
        help="Type to be checked, 'r' for readable, 'w' for writeable.",
    )
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    await async_load_msgdefs(ebus, args)
    type_ = _parse_type(args.type)
    msgdefs = ebus.msgdefs.resolve(args.patterns, filter_=lambda msgdef: _filter_type(msgdef, type_))
    print(f"Listing {msgdefs.summary()}")
    for msgdef in msgdefs:
        for fielddef in msgdef.children:
            details = fielddef.type_.comment
            if fielddef.comment:
                details += f" [{fielddef.comment}]"
            if args.name_only:
                print(fielddef.ident)
            else:
                print(f"{fielddef.ident:<40s} {msgdef.access} {details}")


def _parse_type(type_):
    if type_:
        read, _, write, update = decodetype(type_)
    else:
        read, write, update = None, None, None
    return read, write, update


def _filter_type(msgdef, type_):
    read, write, update = type_
    pairs = ((msgdef.read, read), (msgdef.write, write), (msgdef.update, update))
    return all((exp is None or val == exp) for val, exp in pairs)



"""Observe."""
from .common import (
    add_msgdef_args,
    add_patterns_arg,
    add_read_args,
    async_load_msgdefs,
    create_ebus,
    disable_stdout_buffering,
    print_msg,
)


def parse_args(subparsers):
    """Parse Arguments."""
    parser = subparsers.add_parser(
        "observe",
        help=(
            "Read all known messages once and continue listening so that ALL EBUS values are available, "
            "decode every message and print."
        ),
    )
    add_msgdef_args(parser)
    add_read_args(parser, ttl=10000)
    add_patterns_arg(parser, opt=True)
    parser.set_defaults(main=_main)


async def _main(args):
    disable_stdout_buffering()
    ebus = create_ebus(args)
    await async_load_msgdefs(ebus, args)
    msgdefs = ebus.msgdefs.resolve(args.patterns)
    if args.defaultprio:
        msgdefs.set_defaultprio(args.defaultprio)
    print(f"Observing {msgdefs.summary()}")
    async for msg in ebus.async_observe(msgdefs=msgdefs, ttl=args.ttl):
        print_msg(msg)



"""Cirucit Information Decoder."""
import re

from .circuitinfo import CircuitInfo

_RE_CIRCUIT = re.compile(
    r'^address ([0-9a-f]{2}): .*"MF=([^;]+);ID=([^;]+);SW=([^;]+);HW=([^"]+)", loaded.*".*\.([a-z]+(\.\d+)?)\.csv"'
)

# pylint: disable=C0301


def decode_circuitinfos(info):
    """
    Decode Circuit Information.

    Args:
        info: dict or list of strings.

    >>> info = {
    ...     "version": "ebusd 21.1.v21.1-12-gccfc025",
    ...     "update check": "version 3.4 available",
    ...     "signal": "acquired",
    ...     "symbol rate": "48",
    ...     "max symbol rate": "229",
    ...     "min arbitration micros": "301",
    ...     "max arbitration micros": "3550",
    ...     "min symbol latency": "2",
    ...     "max symbol latency": "15",
    ...     "reconnects": "0",
    ...     "masters": "7",
    ...     "messages": "1006",
    ...     "conditional": "14",
    ...     "poll": "171",
    ...     "update": "10",
    ...     "address 03": "master #11",
    ...     'address 08': 'slave #11, scanned "MF=Vaillant;ID=BAI00;SW=0204;HW=9602", loaded "vaillant/bai.0010015600.inc" ([HW=9602]), "vaillant/08.bai.csv"',
    ...     "address 10": "master #2",
    ...     'address 15': 'slave #2, scanned "MF=Vaillant;ID=UI   ;SW=0508;HW=6201", loaded "vaillant/15.ui.csv"',
    ...     "address 17": "master #17",
    ...     'address 1c': 'slave #17, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/1c.rcc.4.csv"',
    ...     'address 23': 'slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/23.vr630.cc.csv"',
    ...     'address 25': 'slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/25.vr630.hwc.csv"',
    ...     'address 26': 'slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/26.vr630.hc.csv"',
    ...     "address 31": "master #8, ebusd",
    ...     "address 36": "slave #8, ebusd",
    ...     "address 37": "master #18",
    ...     'address 3c': 'slave #18, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/3c.rcc.5.csv"',
    ...     'address 50': 'slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/50.vr630.mc.csv"',
    ...     'address 51': 'slave, scanned "MF=Vaillant;ID=VR630;SW=0500;HW=6301", loaded "vaillant/51.vr630.mc.3.csv"',
    ...     'address 52': 'slave, scanned "MF=Vaillant;ID=MC2  ;SW=0500;HW=6301", loaded "vaillant/52.mc2.mc.4.csv"',
    ...     'address 53': 'slave, scanned "MF=Vaillant;ID=MC2  ;SW=0500;HW=6301", loaded "vaillant/53.mc2.mc.5.csv"',
    ...     "address 70": "master #4",
    ...     'address 75': 'slave #4, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/75.rcc.csv"',
    ...     "address f0": "master #5",
    ...     'address f5': 'slave #5, scanned "MF=Vaillant;ID=RC C ;SW=0508;HW=6201", loaded "vaillant/f5.rcc.3.csv"',
    ... }
    >>> for circuitinfo in decode_circuitinfos(info):
    ...     print(circuitinfo)
    CircuitInfo(circuit='bai', manufacturer='Vaillant', model='BAI00', swversion='0204', hwversion='9602', address=8)
    CircuitInfo(circuit='cc', manufacturer='Vaillant', model='VR630', swversion='0500', hwversion='6301', address=35)
    CircuitInfo(circuit='hc', manufacturer='Vaillant', model='VR630', swversion='0500', hwversion='6301', address=38)
    CircuitInfo(circuit='hwc', manufacturer='Vaillant', model='VR630', swversion='0500', hwversion='6301', address=37)
    CircuitInfo(circuit='mc', manufacturer='Vaillant', model='VR630', swversion='0500', hwversion='6301', address=80)
    CircuitInfo(circuit='mc.3', manufacturer='Vaillant', model='VR630', swversion='0500', hwversion='6301', address=81)
    CircuitInfo(circuit='mc.4', manufacturer='Vaillant', model='MC2', swversion='0500', hwversion='6301', address=82)
    CircuitInfo(circuit='mc.5', manufacturer='Vaillant', model='MC2', swversion='0500', hwversion='6301', address=83)
    CircuitInfo(circuit='rcc', manufacturer='Vaillant', model='RC C', swversion='0508', hwversion='6201', address=117)
    CircuitInfo(circuit='rcc.3', manufacturer='Vaillant', model='RC C', swversion='0508', hwversion='6201', address=245)
    CircuitInfo(circuit='rcc.4', manufacturer='Vaillant', model='RC C', swversion='0508', hwversion='6201', address=28)
    CircuitInfo(circuit='rcc.5', manufacturer='Vaillant', model='RC C', swversion='0508', hwversion='6201', address=60)
    CircuitInfo(circuit='ui', manufacturer='Vaillant', model='UI', swversion='0508', hwversion='6201', address=21)
    """
    circuitinfos = []
    if isinstance(info, dict):
        info = [f"{key}: {value}" for key, value in info.items()]
    for line in info:
        mat = _RE_CIRCUIT.match(line)
        if mat:
            address, manufacturer, model, swversion, hwversion, circuit, _ = mat.groups()
            circuitinfos.append(
                CircuitInfo(
                    circuit.strip(),
                    manufacturer.strip(),
                    model.strip(),
                    swversion.strip(),
                    hwversion.strip(),
                    int(address, 16),
                )
            )
    return tuple(sorted(circuitinfos, key=lambda circuitinfo: circuitinfo.circuit))



"""Utilities."""


def repr_(obj, args=None, kwargs=None):
    """
    Return Python Representation String.

    Keyword Args:
        args (tuple): Tuple with all arg values
        kwargs (tuple): Tuple of (key, value, default) tuples.
    """
    classname = obj.__class__.__qualname__
    args = [repr(arg) for arg in args or []]
    if kwargs:
        for key, value, default in kwargs:
            if value != default:
                args.append("%s=%r" % (key, value))
    return "%s(%s)" % (classname, ", ".join(args))
